from __future__ import annotations

from typing import TYPE_CHECKING, Any

from mknodes.basenodes import mkcontainer
from mknodes.data import eventplugins
from mknodes.utils import classhelpers, inspecthelpers, log, reprhelpers


if TYPE_CHECKING:
    from mkdocs import plugins


logger = log.get_logger(__name__)


class MkPluginFlow(mkcontainer.MkContainer):
    """Node showing info about the different stages of an MkDocs plugin."""

    ICON = "material/dev-to"
    STATUS = "new"

    def __init__(
        self,
        plugin: type[plugins.BasePlugin] | str | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            plugin: MkDocs plugin
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self._plugin = plugin
        self.block_separator = "\n"

    def __repr__(self):
        return reprhelpers.get_repr(self, plugin=self._plugin, _filter_empty=True)

    @property
    def plugins(self):
        match self._plugin:
            case None:
                ep_group = self.event_plugin.entry_point_group
                eps = self.ctx.metadata.entry_points.get(ep_group, [])
                return [i.load() for i in eps]
            case type():
                return [self._plugin]
            case str():
                return [classhelpers.to_class(self._plugin)]
            case _:
                raise TypeError(self._plugin)

    @property
    def event_plugin(self):
        return eventplugins.mkdocs_plugin

    @property
    def items(self):
        import mknodes as mk

        if not self.plugins:
            return []
        items = []
        for plg in self.plugins:
            section = [mk.MkHeader(plg.__name__, parent=self)]
            for event in self.event_plugin.flow:
                if not hasattr(plg, event):
                    continue
                fn = getattr(plg, event)
                code = mk.MkCode.for_object(fn)
                fn_name = fn.__name__
                link = self.event_plugin.help_link.format(event=fn_name)
                link = mk.MkLink(link, fn_name)
                path = self.event_plugin.hook_fn_path.format(event=fn_name)
                info = mk.MkDocStrings(path, show_source=False, show_root_toc_entry=False)
                bubble_content = [
                    mk.MkHeader(link, level=3),
                    mk.MkText(inspecthelpers.get_doc(fn)),
                    mk.MkAdmonition(code, collapsible=True, typ="quote", title="Source"),
                    mk.MkAdmonition(info, collapsible=True, title="Hook info"),
                ]
                bubble = mk.MkSpeechBubble(bubble_content, parent=self)
                section.append(bubble)
            if section and isinstance(section[-1], mk.MkSpeechBubble):
                section[-1].arrow = None
            items.extend(section)
        return items

    @items.setter
    def items(self, value):
        pass

    @classmethod
    def create_example_page(cls, page):
        from mkdocs.contrib import search

        import mknodes as mk

        node = MkPluginFlow(plugin=search.SearchPlugin)
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    node = MkPluginFlow.with_default_context()
    print(node)



from __future__ import annotations

from collections.abc import Mapping

from mknodes.utils import log


logger = log.get_logger(__name__)


STRATEGIES = {
    "additive": [
        (list, "append"),
        (dict, "merge"),
        (set, "union"),
    ],
    "replace": [
        (list, "override"),
        (dict, "override"),
        (set, "override"),
    ],
}


def merge_dicts(
    dct: Mapping,
    *dicts: Mapping,
    strategy: str = "additive",
    deepcopy: bool = False,
):
    import copy

    import deepmerge

    if deepcopy:
        dct = copy.deepcopy(dct)

    strat = STRATEGIES[strategy]
    merger = deepmerge.Merger(strat, ["override"], ["override"])
    for to_merge in dicts:
        dct = merger.merge(dct, to_merge)
    return dct


def merge_extensions(dicts: list[dict[str, dict]]) -> list[dict[str, dict]]:
    seen = set()
    result = []
    dicts = [{k: dct[k]} for dct in dicts for k in dct]
    for dct in dicts:
        dct = dict(sorted(dct.items()))
        if (stringed := str(dct)) not in seen:
            seen.add(stringed)
            result.append(dct)
    return result


if __name__ == "__main__":
    import mknodes as mk

    from mknodes.manual import root

    # dct_a = dict(a=[1, 2, 3], b={"test": "content"})
    # dct_b = dict(a=[1, 2, 3], b={"test2": "content"})
    # print(merge_dicts(dct_a, dct_b, strategy="additive"))
    theme = mk.MaterialTheme()
    proj = mk.Project(theme=theme)
    root.build(proj)



from __future__ import annotations

import functools
import types

from typing import Any, Literal

from mknodes.basenodes import mkdiagram
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


@functools.cache
def get_mermaid(
    package: str | tuple[str] | None = None,
    local_only: bool = True,
    user_only: bool = False,
    include_editables: bool = True,
    editables_only: bool = False,
    reverse: bool = False,
    exclude: str = "",
) -> str:
    """Return mermaid diagram code of dependency tree.

    Mermaid graph code is returned without fences

    Arguments:
        package: package / packages to get a graph for. If None, include all packages
        local_only: Whether to return installs local to the current virtualenv if used
        user_only: If True, only report installation in the user
        include_editables: Whether to include editable installs
        editables_only: Only return editable installs
        reverse: Whether to reverse the graph
        exclude: Packages to exclude from the graph
    """
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        from pip._internal.metadata import pkg_resources
        from pipdeptree._models import PackageDAG
        from pipdeptree._render import render_mermaid

    dists = pkg_resources.Environment.from_paths(None).iter_installed_distributions(
        local_only=local_only,
        skip=(),
        user_only=user_only,
        include_editables=include_editables,
        editables_only=editables_only,
    )
    pkgs = [d._dist for d in dists]  # type: ignore[attr-defined]

    tree = PackageDAG.from_pkgs(pkgs)
    # Reverse the tree (if applicable) before filtering,
    # thus ensuring, that the filter will be applied on ReverseTree
    if reverse:
        tree = tree.reverse()
    match package:
        case str():
            include_list = package.split(",")
        case list() | tuple():
            include_list = list(package)
        case None:
            include_list = None
    exclude_list = set(exclude.split(",")) if exclude else None
    if include_list is not None or exclude_list is not None:
        try:
            tree = tree.filter_nodes(include_list, exclude_list)
        except ValueError:
            msg = "Error when filtering nodes for pipdeptree. Is the package on PyPi?."
            logger.exception(msg)
            return ""
    tree = [tree] if isinstance(tree, str) else tree
    text = render_mermaid(tree)
    return "\n".join(text.splitlines()[1:])


class MkPipDepTree(mkdiagram.MkDiagram):
    """Node to display a mermaid diagram for the dependencies."""

    REQUIRED_PACKAGES = [resources.Package("pipdeptree")]

    def __init__(
        self,
        package: types.ModuleType | str | None = None,
        direction: Literal["TD", "DT", "LR", "RL"] = "TD",
        local_only: bool = False,
        user_only: bool = False,
        include_editables: bool = True,
        editables_only: bool = False,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            package: Package to show a dependency diagram for
            direction: diagram direction
            local_only: Show ony local packages
            user_only: Show only user packages
            include_editables: Whether to include editable installs
            editables_only: Only return editable installs
            kwargs: Keyword arguments passed to parent
        """
        self._package = package
        self.local_only = local_only
        self.user_only = user_only
        self.include_editables = include_editables
        self.editables_only = editables_only
        super().__init__(graph_type="flow", direction=direction, **kwargs)

    def __repr__(self):
        include_editables = self.include_editables if not self.include_editables else None
        return reprhelpers.get_repr(
            self,
            package=self._package,
            local_only=self.local_only,
            user_only=self.user_only,
            include_editables=include_editables,
            editables_only=self.editables_only,
            direction=self.direction,
            _filter_empty=True,
            _filter_false=True,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        tree = MkPipDepTree(direction="LR")
        page += mk.MkReprRawRendered(tree)
        tree = MkPipDepTree("mkdocstrings", direction="LR")
        page += mk.MkReprRawRendered(tree)

    @property
    def package(self) -> str:
        match self._package:
            case None:
                return self.ctx.metadata.distribution_name
            case types.ModuleType():
                return self._package.__name__
            case _:
                return self._package

    @property
    def mermaid_code(self) -> str:
        return get_mermaid(
            tuple(self.package) if isinstance(self.package, list) else self.package,
            local_only=self.local_only,
            user_only=self.user_only,
            include_editables=self.include_editables,
            editables_only=self.editables_only,
        )


if __name__ == "__main__":
    diagram = MkPipDepTree.with_default_context()
    print(diagram)



from __future__ import annotations

from collections.abc import Callable
import os
import pathlib

from typing import Any, get_args

from mknodes import treelib
from mknodes.basenodes import mkcode, mknode
from mknodes.data import treestyles
from mknodes.utils import fsspecpath, log, reprhelpers


logger = log.get_logger(__name__)


class MkTreeView(mkcode.MkCode):
    """Node to display tree structures.

    Currently supports directories and Node subclasses (including `MkNodes`).
    """

    ICON = "material/file-tree-outline"
    STATUS = "new"

    def __init__(
        self,
        tree: str | os.PathLike | treelib.Node,
        *,
        style: treestyles.TreeStyleStr | tuple[str, str, str, str] | None = None,
        maximum_depth: int | None = None,
        predicate: Callable | None = None,
        exclude_folders: list[str] | str | None = None,
        header: str = "",
        storage_options: dict | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            tree: Tree to display. Can be a path to a folder or a Node.
            style: Print style. If tuple, parts are used for stems
            maximum_depth: Maximum nesting depth to print
            predicate: Predicate to filter results
            exclude_folders: Folders to exclude from listing
            header: Section header
            storage_options: Used when tree is a fsspec path
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(header, language="", **kwargs)
        self.tree = tree
        self._style = style
        self.predicate = predicate
        self.maximum_depth = maximum_depth
        self.storage_options = storage_options or {}
        self.exclude_folders = (
            [exclude_folders] if isinstance(exclude_folders, str) else exclude_folders
        )

    @property
    def text(self):
        match self.tree:
            case str() if "://" in self.tree:
                protocol, path = self.tree.split("://", 1)
                node = treelib.FileTreeNode.from_folder(
                    fsspecpath.FsSpecPath(path, protocol, **self.storage_options),
                    predicate=self.predicate,
                    exclude_folders=self.exclude_folders,
                    maximum_depth=self.maximum_depth,
                )
            case str() | os.PathLike():
                node = treelib.FileTreeNode.from_folder(
                    pathlib.Path(self.tree),
                    predicate=self.predicate,
                    exclude_folders=self.exclude_folders,
                    maximum_depth=self.maximum_depth,
                )
            case mknode.MkNode():
                node = self.tree
            case None:
                return ""
            case _:
                raise TypeError(self.tree)
        return node.get_tree_repr(style=self.style, max_depth=self.maximum_depth)

    @text.setter
    def text(self, text):
        self.obj = text

    @property
    def style(self) -> treestyles.TreeStyleStr | tuple[str, str, str, str]:
        return self._style or "rounded"

    @style.setter
    def style(self, value: treestyles.TreeStyleStr | tuple[str, str, str, str] | None):
        self._style = value

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            path=self.tree if isinstance(self.tree, mknode.MkNode) else str(self.tree),
            style=self._style,
            maximum_depth=self.maximum_depth,
            exclude_folders=self.exclude_folders,
            storage_options=self.storage_options,
            _filter_empty=True,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        # Different styles
        for style in get_args(treestyles.TreeStyleStr):
            node = MkTreeView("mknodes/manual", style=style)
            page += mk.MkReprRawRendered(node, header=f"### Style '{style}'")

        # Showing a remote tree structure (using fsspec package)
        # opts = dict(org="mkdocstrings", repo="mkdocstrings")
        # node = MkTreeView("github://", storage_options=opts, maximum_depth=2)
        # page += mk.MkReprRawRendered(node, header="### From remote (FsSpec)")


if __name__ == "__main__":
    opts = dict(org="mkdocstrings", repo="mkdocstrings")
    node = MkTreeView(
        "github://",
        storage_options=opts,
        header="test",
        style="ascii",
        maximum_depth=2,
    )
    print(node.to_markdown())



from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from mknodes.basenodes import mkcontainer, mkimage, mknode
from mknodes.data import badges
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)


class MkShields(mkcontainer.MkContainer):
    """Container for Shields.io / GitHub badges."""

    ICON = "simple/shieldsdotio"

    def __init__(
        self,
        shields: Sequence[badges.BadgeTypeStr] | None = None,
        user: str | None = None,
        project: str | None = None,
        branch: str | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            shields: Shields to include
            user: Github username for shields
            project: project name for shields
            branch: branch name for shields
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.block_separator = "\n"
        self._user = user
        self._project = project
        self._branch = branch
        self._shields = shields

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            shields=self._shields,
            user=self._user,
            project=self._project,
            branch=self._branch,
            _filter_empty=True,
        )

    @property
    def user(self):
        match self._user:
            case None:
                return self.ctx.metadata.repository_username
            case str():
                return self._user

    @user.setter
    def user(self, value):
        self._user = value

    @property
    def project(self):
        if isinstance(self._project, str):
            return self._project
        return self.ctx.metadata.repository_name

    @project.setter
    def project(self, value):
        self._project = value

    @property
    def shields(self):
        return self._shields or [i.identifier for i in badges.SHIELDS]

    @property
    def branch(self) -> str:
        if isinstance(self._branch, str):
            return self._branch
        return self.ctx.git.main_branch or "main"

    @property
    def items(self) -> list[mknode.MkNode]:
        return [
            mkimage.MkImage(
                s.get_image_url(user=self.user, project=self.project, branch=self.branch),
                link=s.get_url(user=self.user, project=self.project),
                title=s.title,
                parent=self,
            )
            for s in badges.SHIELDS
            if s.identifier in self.shields
        ]

    @items.setter
    def items(self, value):
        pass

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkShields(shields=["version", "status", "codecov"])
        page += mk.MkReprRawRendered(node)
        node = MkShields(user="phil65", project="mknodes", shields=None)
        page += mk.MkReprRawRendered(node)
        node = MkShields(user="mkdocs", project="mkdocs", shields=None)
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    shields = MkShields(
        shields=["version", "status", "codecov"],
        user="phil65",
        project="prettyqt",
    )
    print(shields)



from __future__ import annotations

from collections.abc import MutableMapping
import dataclasses
import functools

from typing import Any, Literal

from mknodes.data import datatypes
from mknodes.theme import colortheme, theme
from mknodes.theme.materialtheme import palette
from mknodes.utils import helpers, icons, log, reprhelpers


logger = log.get_logger(__name__)


COLORS = {
    "red": {"color": "#ef5552", "text": "#ffffff"},
    "pink": {"color": "#e92063", "text": "#ffffff"},
    "purple": {"color": "#ab47bd", "text": "#ffffff"},
    "deep purple": {"color": "#7e56c2", "text": "#ffffff"},
    "indigo": {"color": "#4051b5", "text": "#ffffff"},
    "blue": {"color": "#2094f3", "text": "#ffffff"},
    "light blue": {"color": "#02a6f2", "text": "#ffffff"},
    "cyan": {"color": "#00bdd6", "text": "#ffffff"},
    "teal": {"color": "#009485", "text": "#ffffff"},
    "green": {"color": "#4cae4f", "text": "#ffffff"},
    "light green": {"color": "#8bc34b", "text": "#ffffff"},
    "lime": {"color": "#cbdc38", "text": "#000000"},
    "yellow": {"color": "#ffec3d", "text": "#000000"},
    "amber": {"color": "#ffc105", "text": "#000000"},
    "orange": {"color": "#ffa724", "text": "#000000"},
    "deep orange": {"color": "#ff6e42", "text": "#ffffff"},
    "brown": {"color": "#795649", "text": "#ffffff"},
    "grey": {"color": "#757575", "text": "#ffffff"},
    "blue grey": {"color": "#546d78", "text": "#ffffff"},
    "black": {"color": "#000000", "text": "#ffffff"},
    "white": {"color": "#ffffff", "text": "#000000"},
    "custom": {"color": "#4051b5", "text": "#ffffff"},
}


IconTypeStr = Literal[
    "footnotes",
    "details",
    "previous_tab",
    "next_tab",
    "tasklist",
    "tasklist_checked",
]

ICON_TYPE: dict[IconTypeStr, str] = dict(
    footnotes="md-footnotes-icon",
    details="md-details-icon",
    previous_tab="md-tabbed-icon--prev",
    next_tab="md-tabbed-icon--next",
    tasklist="md-tasklist-icon",
    tasklist_checked="md-tasklist-icon--checked",
)


@dataclasses.dataclass
class StatusIcon:
    """Page Status."""

    name: str
    """Slug for referencing the status."""
    svg: str
    """SVG xml for the icon."""
    description: str | None = None
    """Description used for tooltip."""


class MaterialTheme(theme.Theme):
    """Material Theme."""

    name = "material"
    css_template = "css/theme_material.css"

    def __init__(self, **kwargs):
        self._foreground_color = None
        self.classic_admonition_style = True
        self.tooltip_width: int | None = None
        self.content_area_width: int | None = None
        self.default_icons = {}
        self.status_icons = {}
        self.accent_fg_color = None
        self.primary_bg_color = None
        self.color_theme = None
        self.alternate_selector: list[dict[str, str]] = []
        """Layout:
           - name: English
             link: /en/
             lang: en
        """
        super().__init__(self.name, **kwargs)

    def get_css_context(self) -> dict[str, Any]:
        """Return template context (used to render the CSS template."""
        return dict(
            admonitions=self.admonitions,
            classic_admonition_style=self.classic_admonition_style,
            alternating_table_colors=self.alternating_table_colors,
            tooltip_width=self.tooltip_width,
            content_area_width=self.content_area_width,
            default_icons=self.default_icons,
            status_icons=list(self.status_icons.values()),
            accent_fg_color=self.accent_fg_color,
            primary_bg_color=self.primary_bg_color,
            color_theme=self.color_theme,
            css_primary_fg="var(--md-primary-fg-color)",
            css_primary_bg="var(--md-primary-bg-color)",
            css_primary_bg_light="var(--md-primary-bg-color--light)",
            css_accent_fg="var(--md-accent-fg-color)",
            css_accent_fg_transparent="var(--md-accent-fg-color--transparent)",
            css_accent_bg="var(--md-accent-bg-color)",
            css_default_fg="var(--md-code-fg-color)",
            css_default_bg="var(--md-code-bg-color)",
        )

    def __repr__(self):
        return reprhelpers.get_repr(self)

    @functools.cached_property
    def palettes(self) -> list[palette.Palette]:
        """Return a list of palettes used by the theme."""
        data = self.data.get("palette")
        if not data:
            return [palette.Palette()]
        if isinstance(data, dict):
            data = [data]
        return [
            palette.Palette(
                primary=pal.get("primary", "indigo"),
                accent=pal.get("accent", "indigo"),
                scheme=pal.get("scheme", "default"),
                media=pal.get("media"),
                toggle_name=pal.get("toggle", {}).get("name"),
                toggle_icon=pal.get("toggle", {}).get("icon"),
            )
            for pal in data
        ]

    def get_accent_color(self) -> str:
        """Get the accent foreground color."""
        # sourcery skip: use-or-for-fallback
        color = self._get_color("accent", fallback="")
        if not color:
            color = self._get_color("primary", fallback="indigo")
        return COLORS[color]["color"]

    def _get_color(self, color_type: Literal["primary", "accent"], fallback: str) -> str:
        """Return either primary or accent color.

        Arguments:
            color_type: The color type to get
            fallback: value to return in case color_type is not defined.
        """
        pal = self.palettes[0]
        match color_type:
            case "primary":
                return pal.primary or fallback
            case "accent":
                return pal.accent or fallback

    def _set_color(self, color_type: Literal["primary", "accent"], value: str):
        """Set the color for the first palette.

        Arguments:
            color_type: The color type to set
            value: The color to set
        """
        pal = self.palettes[0]
        match color_type:
            case "primary":
                pal.primary = value
            case "accent":
                pal.accent = value

    def set_accent_foreground_color(self, color: datatypes.RGBColorType):
        """Set the accent foreground color.

        Arguments:
            color: Color to set
        """
        color_str = helpers.get_color_str(color)
        self.accent_fg_color = color_str
        self._set_color("accent", "custom")
        return color_str

    def set_primary_background_color(
        self,
        color: datatypes.RGBColorType,
    ):
        """Set the primary background color.

        Arguments:
            color: Color to set
        """
        self._set_color("primary", "custom")
        color_str = helpers.get_color_str(color)
        self.primary_bg_color = color_str
        return color_str

    @property
    def primary_color(self) -> str:
        """Get the primary foreground color."""
        if self._foreground_color:
            return self._foreground_color
        color = self._get_color("primary", fallback="indigo")
        return COLORS[color]["color"]

    @property
    def _text_color(self) -> str:
        """Get the primary text color."""
        color = self._get_color("primary", fallback="indigo")
        return COLORS[color]["text"]

    def set_primary_foreground_color(
        self,
        color: datatypes.ColorType,
        light_shade: datatypes.ColorType | None = None,
        dark_shade: datatypes.ColorType | None = None,
    ):
        """Set a custom color theme.

        Requires primary color set to "custom".

        Arguments:
            color: Main color
            light_shade: Optional light shade. If None, same as color.
            dark_shade: Optional dark shade. If None, same as color.
        """
        self._foreground_color = color
        self._set_color("primary", "custom")
        if light_shade is None:
            light_shade = color
        if dark_shade is None:
            dark_shade = color
        color_str = helpers.get_color_str(color)
        self.color_theme = colortheme.ColorTheme(
            color_str,
            helpers.get_color_str(light_shade),
            helpers.get_color_str(dark_shade),
        )
        return color_str

    def set_default_icon(self, icon_type: IconTypeStr, data: str):
        """Allows setting some custom default icons used throughout the theme.

        Arguments:
            icon_type: Icon type to set a new default for
            data: svg for the new icon
        """
        self.default_icons[icon_type] = data

    def add_status_icon(
        self,
        name: str,
        material_icon: str,
        description: str | None = None,
    ):
        """Add a custom status icon.

        Arguments:
            name: slug for the status icon
            material_icon: Material icon name
            description: Optional status description (used for tooltip)
        """
        data = icons.get_icon_svg(material_icon)
        self.status_icons[name] = StatusIcon(name, data, description)

    def adapt_extensions(self, extensions: MutableMapping[str, dict]):
        """MkDocs-Material needs some custom configuration for extensions.

        This method will get called during the build process in order to make sure
        that theme-specifics are considered.

        Arguments:
            extensions: Extensions to adapt.
        """
        for k in dict(extensions).copy():
            ext = extensions[k]
            if k in ["pymdownx.blocks.tab", "pymdownx.tabbed"]:
                ext["alternate_style"] = True
            elif k == "pymdownx.tasklist":
                ext["custom_checkbox"] = True

    def adapt_extras(self, extras: dict):
        if self.alternate_selector:
            extras.setdefault("alternate", []).extend(self.alternate_selector)
        for status in self.status_icons.values():
            if status.description:
                extras.setdefault("status", {})[status.name] = status.description


if __name__ == "__main__":
    theme = MaterialTheme()



from __future__ import annotations

from mknodes.basenodes import mkcontainer, mknode
from mknodes.utils import log, reprhelpers, xmlhelpers as xml


logger = log.get_logger(__name__)


class MkGridCard(mkcontainer.MkContainer):
    """Node representing a single grid card."""

    ICON = "material/square-medium"
    STATUS = "new"

    def __init__(
        self,
        content: list | str | mknode.MkNode | None = None,
        **kwargs,
    ):
        super().__init__(content=content, **kwargs)

    def _to_markdown(self) -> str:
        item_str = "\n".join(i.to_markdown() for i in self.items)
        lines = item_str.split("\n")
        result = [f"-   {lines[0]}"]
        result.extend(f"    {i}" for i in lines[1:])
        return "\n".join(result) + "\n"


class MkGrid(mkcontainer.MkContainer):
    """Node for showing a grid."""

    ICON = "material/view-grid"
    items: list[MkGridCard]

    def __init__(
        self,
        cards: list[str | mknode.MkNode] | None = None,
        *,
        header: str = "",
        **kwargs,
    ):
        match cards:
            case None:
                items = []
            case list():
                items = [
                    (card if isinstance(card, MkGridCard) else MkGridCard(card))
                    for card in cards
                ]
        super().__init__(content=items, header=header, **kwargs)

    def __repr__(self):
        cards = []
        for item in self.items:
            if len(item.items) == 1:
                cards.append(reprhelpers.to_str_if_textnode(item.items[0]))
            else:
                cards.append(item)
        return reprhelpers.get_repr(self, cards=cards)

    def get_element(self) -> xml.Div:
        root = xml.Div("grid cards", markdown=True)
        content = "".join(str(i) for i in self.items) if self.items else ""
        root.text = "\n\n" + content + "\n"
        return root

    def _to_markdown(self) -> str:
        root = self.get_element()
        return root.to_string(space="")

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        # only works for Mkdocs-material sponsors.
        ls = ["Item 1", "Item 2", "Item 3"]
        item_1 = mk.MkList(ls)
        item_2 = mk.MkKeys(keys=["Ctrl+A"])
        grid = MkGrid(cards=[item_1, item_2])
        page += mk.MkReprRawRendered(grid)


if __name__ == "__main__":
    import mknodes as mk

    keys = mk.MkKeys(keys="Ctrl+A")
    grid = MkGrid(cards=[keys, keys])
    print(grid)



from __future__ import annotations

import os
import pathlib

from mknodes.utils import downloadhelpers, helpers, superdict


class ConfigFile(superdict.SuperDict):
    def __init__(self, path: str | os.PathLike | None = None):
        """Constructor.

        Arguments:
            path: Path to the config file
        """
        super().__init__()
        self.path = path
        if self.path is None:
            return
        self.path = str(path)
        if helpers.is_url(self.path):
            content = downloadhelpers.download(self.path)
            self.load_config(content.decode())
        else:
            self.load_file(self.path)

    def __repr__(self):
        return f"{type(self).__name__}({self.path!r})"

    def __bool__(self):
        return bool(self._data or self.path)

    def get_section_text(
        self,
        *sections: str,
        keep_path: bool = False,
    ) -> str:
        """Try to get data from given path as text.

        If Key path does not exist, return empty string.

        Arguments:
            sections: Sections to dig into
            keep_path: Return result with original nesting
        """
        if not sections:
            raise ValueError(sections)
        section = self.get_section(*sections, keep_path=keep_path)
        return "" if section is None else self._dump(section)

    def load_config(self, data: str):
        """Load a string with loader of given file type.

        Arguments:
            data: String with markup of type as config file
        """
        self._data = self._load(data)

    def dump_config(self) -> str:
        """Dump to string with dumper of given file type."""
        return self._dump(self._data)

    def load_file(self, path: str | os.PathLike):
        """Load a file with loader of given file type.

        Arguments:
            path: Path to the config file
        """
        text = pathlib.Path(path).read_text(encoding="utf-8")
        self.load_config(text)

    @classmethod
    def _dump(cls, data: dict) -> str:
        """Needs to be reimplemented by subclasses.

        Arguments:
            data: Data to dump
        """
        raise NotImplementedError

    @classmethod
    def _load(cls, data: str) -> dict | list:
        """Needs to be reimplemented by subclasses.

        Arguments:
            data: Data to load
        """
        raise NotImplementedError


if __name__ == "__main__":
    from mknodes.info import tomlfile

    info = tomlfile.TomlFile("pyproject.toml")
    text = info.get_section_text("tool", "hatch", keep_path=True)
    print(text)



from __future__ import annotations

import collections
import functools
import os
import pathlib

from typing import Any

from mknodes.data import buildsystems, commitconventions, installmethods
from mknodes.info import tomlfile
from mknodes.utils import pathhelpers, superdict


class PyProject(tomlfile.TomlFile):
    """Class representing a PyProject config file."""

    def __init__(self, path: str | os.PathLike | None = None):
        """Constructor.

        Arguments:
            path: Path to the pyproject file.
                  If None, parent directories are checked, too.
                  If path points to folder, check that folder for a pyproject.toml
                  Otherwise, take file from explicit path.
        """
        if path is None:
            path = pathhelpers.find_file_in_folder_or_parent("pyproject.toml")
        if path is None:
            msg = "Could not find pyproject.toml"
            raise FileNotFoundError(msg)
        path = pathlib.Path(path)
        if path.is_dir():
            path = path / "pyproject.toml"
        super().__init__(path)

    @property
    def mknodes_section(self) -> dict[str, Any]:
        """Return our very own config section."""
        return self.get_section("tool", "mknodes") or {}

    @property
    def name(self) -> str | None:
        """Project name."""
        return self.project.get("name")

    @property
    def tool(self) -> superdict.SuperDict[Any]:
        """Tool section."""
        dct: collections.defaultdict[str, dict] = collections.defaultdict(dict)
        dct.update(self._data.get("tool", {}))
        return superdict.SuperDict(dct)

    @property
    def project(self) -> dict[str, Any]:
        """Project section."""
        return self._data.get("project", {})

    @functools.cached_property
    def configured_build_systems(self) -> list[buildsystems.BuildSystemStr]:
        """Return build systems which have a config in tools section."""
        return [p for p in buildsystems.BUILD_SYSTEMS if p in self.tool]

    @functools.cached_property
    def build_system(self) -> buildsystems.BuildSystem:
        """Return the build system set as build backend."""
        back_end = self._data["build-system"]["build-backend"]
        for p in buildsystems.BUILD_SYSTEMS.values():
            if p.build_backend == back_end:
                return p
        msg = "No known build backend"
        raise RuntimeError(msg)

    @property
    def allowed_commit_types(self) -> list[commitconventions.CommitTypeStr]:
        """Return the allowed commit types."""
        return self.mknodes_section.get("allowed-commit-types", [])

    @property
    def extras_descriptions(self) -> dict[str, str]:
        """Return a dictionary with descriptions for dependency extras."""
        return self.mknodes_section.get("extras-descriptions", {})

    @property
    def package_repos(self) -> list[installmethods.InstallMethodStr]:
        """Return a list of package repositories the package is available on."""
        return self.mknodes_section.get("package-repositories", [])

    @property
    def docstring_style(self) -> str | None:
        """Return the style used for docstring."""
        if convention := self.tool["pydocstyle"].get("convention"):
            return convention
        return self.mknodes_section.get("docstring-style")

    @property
    def argparser_path(self) -> str | None:
        """Return path to an ArgumentParser, used for documenting the CLI."""
        return self.mknodes_section.get("argparser-path")

    @property
    def line_length(self) -> int | None:
        # sourcery skip: assign-if-exp, reintroduce-else
        """Return the line length (taken from black / ruff / isort config)."""
        if length := self.tool["ruff"].get("line-length"):
            return int(length)
        if length := self.tool["isort"].get("line_length"):
            return int(length)
        if length := self.tool["black"].get("line-length"):
            return int(length)
        return None

    @property
    def min_tool_python_version(self) -> tuple[int, int] | None:
        """Return python target version of the tools.

        Taken from black / ruff / mypy config.
        """
        if version := self.tool["mypy"].get("python_version"):
            return (3, int(version[-2:]))
        if version := self.tool["black"].get("target-version"):
            return (3, int(version[-2:]))
        if version := self.tool["ruff"].get("target-version"):
            return (3, int(version[-2:]))
        return None

    # @functools.cached_property
    # def context(self):
    #     return contexts.PyProjectContext(
    #         extras_descriptions=self.extras_descriptions,
    #         configured_build_systems=self.configured_build_systems,
    #         build_system=self.build_system,
    #         package_repos=self.package_repos,
    #         commit_types=self.allowed_commit_types,
    #         docstring_style=self.docstring_style,
    #         line_length=self.line_length,
    #         tool_section=self.tool,
    #     )


if __name__ == "__main__":
    info = PyProject()
    print(info.min_tool_python_version)



from __future__ import annotations

from collections.abc import Callable
import os
import types

from typing import Any, Literal

from mknodes.basenodes import mknode
from mknodes.utils import classhelpers, log, reprhelpers, resources


logger = log.get_logger(__name__)


class MkDocStrings(mknode.MkNode):
    """Docstring section (powered by mkdocstrings)."""

    REQUIRED_PLUGINS = [resources.Plugin("mkdocstrings")]
    REQUIRED_PACKAGES = [resources.Package("mkdocstrings", extras=["python"])]
    OPTIONS_DEFAULT: dict[str, Any] = {}
    ICON = "material/api"

    def __init__(
        self,
        obj: (
            types.ModuleType
            | str
            | tuple[str, ...]
            | list[str]
            | os.PathLike
            | type
            | Callable
        ),
        for_topmost: bool = True,
        allow_inspection: bool | None = None,
        show_bases: bool | None = None,
        show_source: bool | None = None,
        preload_modules: list[str] | None = None,
        heading_level: int | None = None,
        show_root_heading: bool | None = None,
        show_root_toc_entry: bool | None = None,
        show_root_full_path: bool | None = None,
        show_root_members_full_path: bool | None = None,
        show_object_full_path: bool | None = None,
        show_category_heading: bool | None = None,
        show_symbol_type_heading: bool | None = None,
        show_symbol_type_toc: bool | None = None,
        inherited_members: bool | None = None,
        members: list[str] | None = None,
        members_order: Literal["alphabetical", "source"] | None = None,
        filters: list[str] | None = None,
        group_by_category: bool | None = None,
        show_submodules: bool | None = None,
        docstring_section_style: Literal["table", "list", "spacy"] | None = None,
        merge_init_into_class: bool | None = None,
        show_if_no_docstring: bool | None = None,
        annotations_path: Literal["brief", "source"] | None = None,
        line_length: int | None = None,
        show_signature: bool | None = None,
        show_signature_annotations: bool | None = None,
        signature_crossrefs: bool | None = None,
        separate_signature: bool | None = None,
        **kwargs: Any,
    ):
        """Docstring section.

        Global options for DocStrings can be overridden by setting the keyword arguments
        to not-None.

        Arguments:
            obj: What to show DocStrings for.
            for_topmost: If True, try to find the "shortest" path to given object by
                         checking whether it can also be found in a parent module.
            allow_inspection: Whether to allow inspecting modules when visiting
                              them is not possible
            show_bases: Show the base classes of a class.
            show_source: Show the source code of this object.
            preload_modules: List of modules to pre-load.
            heading_level: The initial heading level to use.
            show_root_heading: Show the heading of the object at the root of the
                               documentation tree (i.e. the object referenced by
                               the identifier after :::).
            show_root_toc_entry: If the root heading is not shown, at least
                                 add a ToC entry for it.
            show_root_full_path: Show the full Python path for the root
                                 object heading.
            show_root_members_full_path: Show the full Python path of the
                                         root members.
            show_object_full_path: Show the full Python path of every object.
            show_category_heading: When grouped by categories, show a heading
                                   for each category.
            show_symbol_type_heading: Show the symbol type in headings (e.g. mod,
                                      class, func and attr).
            show_symbol_type_toc: Show the symbol type in the Table of
                                  Contents (e.g. mod, class, func and attr).
            inherited_members: Also show inherited members.
            members: An explicit list of members to render.
            members_order: The members ordering to use.
            filters: A list of filters applied to filter objects based on their name.
                     A filter starting with ! will exclude matching objects instead of
                     including them. The members option takes precedence over filters
                     (filters will still be applied recursively to lower members in the
                     hierarchy).
            group_by_category: Group the object's children by categories:
                               attributes, classes, functions, and modules.
            show_submodules: When rendering a module, show its submodules recursively.
            docstring_section_style: The style used to render docstring sections.
            merge_init_into_class: Whether to merge the __init__ method into
                                   the class' signature and docstring.
            show_if_no_docstring: Show the object heading even if it has no
                                  docstring or children with docstrings.
            annotations_path: The verbosity for annotations path
            line_length: Maximum line length when formatting code/signatures.
            show_signature: Show methods and functions signatures.
            show_signature_annotations: Show the type annotations in methods
                                        and functions signatures.
            signature_crossrefs: Whether to render cross-references for type
                                 annotations in signatures.
            separate_signature: Whether to put the whole signature in a code
                                block below the heading. If Black is installed,
                                the signature is also formatted using it.
            kwargs: Keyword arguments passed to super.
        """
        super().__init__(**kwargs)
        self.obj = obj
        self.for_topmost = for_topmost
        opts = dict(
            allow_inspection=allow_inspection,
            show_bases=show_bases,
            show_source=show_source,
            preload_modules=preload_modules,
            heading_level=heading_level,
            show_root_heading=show_root_heading,
            show_root_toc_entry=show_root_toc_entry,
            show_root_full_path=show_root_full_path,
            show_root_members_full_path=show_root_members_full_path,
            show_object_full_path=show_object_full_path,
            show_category_heading=show_category_heading,
            show_symbol_type_heading=show_symbol_type_heading,
            show_symbol_type_toc=show_symbol_type_toc,
            inherited_members=inherited_members,
            members=members,
            members_order=members_order,
            filters=filters,
            group_by_category=group_by_category,
            show_submodules=show_submodules,
            docstring_section_style=docstring_section_style,
            merge_init_into_class=merge_init_into_class,
            show_if_no_docstring=show_if_no_docstring,
            annotations_path=annotations_path,
            line_length=line_length,
            show_signature=show_signature,
            show_signature_annotations=show_signature_annotations,
            signature_crossrefs=signature_crossrefs,
            separate_signature=separate_signature,
        )
        opts = self.OPTIONS_DEFAULT.copy() | opts
        self._options = {k: v for k, v in opts.items() if v is not None}

    def __repr__(self):
        option_kwargs = {k: v for k, v in self.options.items() if v is not None}
        return reprhelpers.get_repr(self, obj=self.obj, **option_kwargs)

    @property
    def obj_path(self):
        match self.obj:
            case types.ModuleType():
                return self.obj.__name__
            case Callable():
                if not self.for_topmost:
                    return f"{self.obj.__module__}.{self.obj.__qualname__}"
                topmost_path = classhelpers.get_topmost_module_path(self.obj)
                return f"{topmost_path}.{self.obj.__qualname__}"
            case str():
                return self.obj  # for setting a manual path
            case tuple() | list():
                return ".".join(self.obj)
            case _:
                raise TypeError(self.obj)

    @property
    def options(self):
        import mknodes as mk

        # The default section style does not work well inside Annotations,
        # so we switch to "list" when any parent is an MkAnnotation
        # and no style is explicitely set.
        opts = self._options.copy()
        style = opts.get("docstring_section_style")
        if not style and self.is_descendant_of(mk.MkAnnotations):
            opts["docstring_section_style"] = "list"
            opts["show_root_heading"] = True
        return opts

    def _to_markdown(self) -> str:
        md = f"::: {self.obj_path}\n"
        if not self.options:
            return md
        option_lines = [f"      {k}: {v!r}" for k, v in self.options.items()]
        option_text = "\n".join(option_lines)
        return f"{md}    options:\n{option_text}\n"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        page += "The MkDocStrings node shows DocStrings from mkdocstrings addon."
        node = MkDocStrings(
            "mknodes.MkDocStrings",
            show_if_no_docstring=True,
            header="DocStrings",
            heading_level=3,
        )
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    docstrings = MkDocStrings("a.b", show_submodules=True)
    print(docstrings)



from __future__ import annotations

from mknodes.utils import inventorymanager


DOTTED_PATH = "prettyqt.widgets.widget.WidgetMixin.set_style"
BASE_URL = "http://test.de"
EXPECTED = f"http://test.de/qt_modules/widgets/WidgetMixin.html#{DOTTED_PATH}"


def test_getting_link(test_data_dir):
    inv_manager = inventorymanager.InventoryManager()
    inv_manager.add_inv_file(test_data_dir / "objects.inv", base_url=BASE_URL)
    item = inv_manager[DOTTED_PATH]
    assert item == EXPECTED



from __future__ import annotations

import textwrap

from typing import TYPE_CHECKING

from mknodes.basenodes import mkcard, mkcontainer, mknode
from mknodes.pages import mkpage
from mknodes.utils import helpers, log, reprhelpers


if TYPE_CHECKING:
    import mknodes as mk


logger = log.get_logger(__name__)


class MkShowcase(mkcontainer.MkContainer):
    """Node for showing a html-based image grid.

    Manages row / column positioning.
    Mainly intended for MkCards, but can also include other markdown (there are limits
    though.)
    When adding MkCards, then addtional CSS is required.
    """

    ICON = "material/view-grid"

    def __init__(
        self,
        cards: list[str | mknode.MkNode] | None = None,
        column_count: int = 3,
        *,
        header: str = "",
        **kwargs,
    ):
        self.column_count = column_count
        super().__init__(content=cards or [], header=header, **kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(self, cards=self.items)

    def to_child_node(self, item) -> mknode.MkNode:
        match item:
            case mkpage.MkPage():
                return mkcard.MkCard(
                    target=item,
                    title=item.title or " ",
                    caption=item.subtitle or " ",
                    image=":material-tab:",
                    parent=self,
                )
            case mkcard.MkCard():
                return item
            case _:
                return super().to_child_node(item)

    def _to_markdown(self) -> str:
        text = ""
        for items in helpers.batched(self.items, self.column_count):
            text += '<div class="row">'
            for item in items:
                text += '\n  <div class="column">\n'
                text += textwrap.indent(str(item), "    ")
                text += "  </div>"
            text += "\n</div>"
        return text

    def add_card(
        self,
        title: str,
        image: str,
        link: str | mk.MkPage | mk.MkNav | None = None,
        caption: str | None = None,
    ):
        """Add an image card to the node.

        Arguments:
            title: Card title
            image: link to the Image
            link: Optional link for the card
            caption: Image caption
        """
        card = mkcard.MkCard(target=link, title=title, image=image, caption=caption)
        self.append(card)

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkShowcase()
        for i in range(9):
            node.add_card(
                link="https://phil65.github.io/mknodes/",
                title=f"Title {i}",
                image="https://picsum.photos/300",
                caption=f"Caption {i}",
            )
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    import mknodes as mk

    keys = mk.MkKeys(keys="Ctrl+A")
    grid = MkShowcase()
    grid.add_card("Tse", "td", "http://www.google.com")
    print(grid.items[0])



from __future__ import annotations

from typing import Any, Literal

from mknodes.basenodes import mkblock, mknode
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


MarkdownModeStr = Literal["block", "inline", "raw", "auto", "html"]

RawBlockTags = Literal["pre", "canvas", "math", "option"]
RawHtmlTags = Literal["script", "style"]
BlockTypeStr = Literal["div", "span", "code"] | RawHtmlTags | RawBlockTags


class MkHtmlBlock(mkblock.MkBlock):
    """PyMdown-based Html block. Can be used to show raw content."""

    ICON = "octicons/code-16"
    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.blocks.html")]

    def __init__(
        self,
        content: str | mknode.MkNode = "",
        *,
        block_type: BlockTypeStr = "div",
        markdown_mode: MarkdownModeStr | None = None,
        attributes: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            content: What should be contained in the block
            block_type: Block type
            markdown_mode: Markdown mode
            attributes: Block attrs
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(
            name="html",
            argument=block_type,
            content=content,
            attributes=attributes,
            **kwargs,
        )
        self.markdown_mode = markdown_mode
        self.indent = "    "

    def __repr__(self):
        if len(self.items) == 1:
            content = reprhelpers.to_str_if_textnode(self.items[0])
        else:
            content = [reprhelpers.to_str_if_textnode(i) for i in self.items]
        return reprhelpers.get_repr(
            self,
            content=content,
            block_type=self.argument,
            markdown_mode=self.markdown_mode,
            attributes=self.attributes,
            _filter_empty=True,
        )

    @property
    def markdown_mode(self) -> MarkdownModeStr | None:
        """The markdown mode attribute."""
        return self.attributes.get("markdown")

    @markdown_mode.setter
    def markdown_mode(self, value: MarkdownModeStr | None):
        self.attributes["markdown"] = value


if __name__ == "__main__":
    test = MkHtmlBlock("test")
    print(test)



from __future__ import annotations

import dataclasses


# from typing import Literal


@dataclasses.dataclass
class AdmonitionType:
    name: str
    svg: str
    header_color: str
    icon_color: str
    border_color: str
    font_color: str



from __future__ import annotations

from abc import ABCMeta
from collections.abc import Iterator, MutableMapping
import os

from typing import Any, Literal, TypeVar

from mknodes.utils import pathhelpers


MarkupTypeStr = Literal["yaml", "json", "toml"]


V = TypeVar("V")


class SuperDict(MutableMapping[str, V], metaclass=ABCMeta):
    def __init__(self, data: dict | None = None, **kwargs):
        self._data: dict[str, V] = data or {}
        self._data |= kwargs

    def __getitem__(self, value: str) -> V:
        if isinstance(value, str):
            return self._data.__getitem__(value)
        return self.get_section(value)

    def __setitem__(self, index: str, value: V):
        self._data[index] = value

    def __delitem__(self, index: str):
        del self._data[index]

    def __bool__(self):
        return bool(self._data)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data.keys())

    def __len__(self):
        return len(self._data)

    def get_section(self, *sections: str, keep_path: bool = False) -> Any:
        """Try to get data with given section path.

        If Key path does not exist, return None.

        Arguments:
            sections: Sections to dig into
            keep_path: Return result with original nesting
        """
        section = self._data
        for i in sections:
            if isinstance(section, dict):
                if child := section.get(i):
                    section = child
                else:
                    return None
            else:
                for idx in section:
                    if i in idx and isinstance(idx, dict):
                        section = idx[i]
                        break
                    if isinstance(idx, str) and idx == i:
                        section = idx
                        break
                else:
                    return None
        if not keep_path:
            return SuperDict(section) if isinstance(section, dict) else section
        result: dict[str, dict] = {}
        new = result
        for sect in sections:
            result[sect] = section if sect == sections[-1] else {}
            result = result[sect]
        return SuperDict(new) if isinstance(new, dict) else new

    def serialize(self, mode: MarkupTypeStr | None) -> str:  # type: ignore[return]
        match mode:
            case None | "yaml":
                from mknodes.utils import yamlhelpers

                return yamlhelpers.dump_yaml(self._data)
            case "json":
                import json

                return json.dumps(self._data, indent=4)
            case "toml" if isinstance(self._data, dict):
                import tomli_w

                return tomli_w.dumps(self._data)
            case _:
                raise TypeError(mode)

    def write(self, path: str | os.PathLike, mode: MarkupTypeStr | None = None):
        text = self.serialize(mode)
        pathhelpers.write_file(text, path)


if __name__ == "__main__":
    dct = SuperDict(a=dict(b="c"))  # type: ignore[var-annotated]
    text = dct.get_section("a", "b", keep_path=False)
    print(text)



from __future__ import annotations

import datetime
import functools
import os

import github

from github import Commit

from mknodes.info import contexts
from mknodes.utils import downloadhelpers, log, pathhelpers, reprhelpers


RAW_URL = "https://raw.githubusercontent.com/"
TOKEN = os.environ.get("GITHUB_TOKEN")

auth = github.Auth.Token(TOKEN) if TOKEN else None


logger = log.get_logger(__name__)


class GitHubRepo:
    def __init__(self, username: str, repository: str):
        """Constructor.

        Arguments:
            username: Github user name
            repository: Github repository / project name
        """
        self.main = github.Github(auth=auth)
        self.username = username
        self.repo_name = repository
        self.user: github.NamedUser.NamedUser = self.main.get_user(username)  # type: ignore
        self.repo = self.main.get_repo(f"{username}/{repository}")
        self.default_branch = self.repo.default_branch

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            username=self.username,
            repository=self.repo_name,
        )

    def download_from_path(
        self,
        path: str | os.PathLike,
        destination: str | os.PathLike,
        recursive: bool = False,
    ):
        """Download a file from this github repository.

        Arguments:
            path: Path to the file we want to download.
            destination: Path where file should be saved.
            recursive: Download all files from a folder (and subfolders).
        """
        user_name = self.main.get_user().login if TOKEN else None
        return pathhelpers.download_from_github(
            org=self.username,
            repo=self.repo_name,
            path=path,
            destination=destination,
            username=user_name,
            token=TOKEN,
            recursive=recursive,
        )

    @functools.cached_property
    def workflows(self) -> list[dict[str, str]]:
        """Return a list of dictionaries containing info about the current workflows."""
        result = []
        for wf in self.repo.get_workflows():
            url = f"{self.raw_prefix}{self.default_branch}/{wf.path}"
            data = downloadhelpers.download(url)
            item = dict(name=wf.name, workflow=data.decode(), badge_url=wf.badge_url)
            result.append(item)
        return result

    @functools.cached_property
    def raw_prefix(self) -> str:
        """Path to the base url of the "Raw" links of GitHub.

        If used in combination with the relative path of a module file,
        this geives a valid link to a GitHub raw link.
        """
        return f"{RAW_URL}{self.username}/{self.repo_name}/"

    def get_last_commits(
        self,
        branch: str | None = None,
        since: datetime.datetime | None = None,
    ) -> list[Commit.Commit]:  # type: ignore[valid-type]
        """Return last x commits.

        Arguments:
            since: Amount of commits to fetch.
            branch: Branch to get commits from. Defaults to main / master.
        """
        rev = branch or self.default_branch
        kwargs = dict(since=since) if since else {}
        return self.repo.get_commits(rev, **kwargs)  # type: ignore

    @functools.cached_property
    def context(self):
        """Return github context."""
        return contexts.GitHubContext(
            default_branch=self.default_branch,
            repo_name=self.repo_name,
            workflows=self.workflows,
            avatar_url=self.user.avatar_url,
            bio=self.user.bio,
            blog=self.user.blog,
            company=self.user.company,
            # contributions=self.user.contributions,
            email=self.user.email,
            followers=self.user.followers,
            gravatar_id=self.user.gravatar_id,
            hireable=self.user.hireable,
            location=self.user.location,
            name=self.user.name,
            twitter_username=self.user.twitter_username,
        )


if __name__ == "__main__":
    g = GitHubRepo("phil65", "mknodes")
    import rich

    node = g.main.get_user()
    rich.inspect(node)



from __future__ import annotations

from typing import Any

from mknodes.basenodes import mktext
from mknodes.info import license
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)


class MkLicense(mktext.MkText):
    """Node to show a license.

    If not explicitely set, the license will be pulled from the project.

    """

    ICON = "material/license"
    STATUS = "new"

    def __init__(
        self,
        license_type: str | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            license_type: License to show (identifier from https://spdx.org/licenses/)
                          If none is set, it will try to get license from Project
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.license = license_type

    def __repr__(self):
        return reprhelpers.get_repr(self, license=self.license, _filter_empty=True)

    @property
    def text(self):
        if self.license:
            lic = license.License.from_name(self.license)
            lic.resolve_by_distribution(self.ctx.metadata.distribution_name)
            return lic.content
        return self.ctx.metadata.license_text or ""

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkLicense()
        page += mk.MkReprRawRendered(node, header="### From project")
        node = MkLicense("GPL-3.0")
        page += mk.MkReprRawRendered(node, header="### Explicit")


if __name__ == "__main__":
    lic = MkLicense.with_default_context("GPL-3.0")
    print(lic.text)



from __future__ import annotations

from abc import ABCMeta
from collections.abc import MutableMapping
import os
import tempfile

from mknodes.info import gitrepository
from mknodes.utils import helpers, log


logger = log.get_logger(__name__)


def get_repo(
    repo_url: str | os.PathLike,
    clone_depth: int = 100,
) -> gitrepository.GitRepository:
    """Return info for given module from registry.

    Arguments:
        repo_url: Name of the module
        clone_depth: Amount of commits to clone if repo is remote.
    """
    return registry.get_repo(repo_url, clone_depth=clone_depth)


class RepoRegistry(MutableMapping, metaclass=ABCMeta):
    """Registry for Git repositories.

    Manages the GitRepository instances and temporary directories.
    """

    def __init__(self):
        self._repos: dict[str, gitrepository.GitRepository] = {}
        self._directories: dict[str, tempfile.TemporaryDirectory] = {}

    def __getitem__(self, value):
        return self._repos.__getitem__(value)

    def __setitem__(self, index, value):
        self._repos[index] = value

    def __delitem__(self, index):
        del self._repos[index]

    def __repr__(self):
        return f"{type(self).__name__}()"

    def __iter__(self):
        return iter(self._repos.keys())

    def __len__(self):
        return len(self._repos)

    def get_repo(
        self,
        repo_url: str | os.PathLike,
        clone_depth: int = 100,
    ) -> gitrepository.GitRepository:
        """Return a GitRepository for given URL / path.

        Arguments:
            repo_url: URL / path of the repository
            clone_depth: Amount of commits to fetch if repo is remote
        """
        repo_url = os.fspath(repo_url)
        if repo_url in self._repos:
            return self._repos[repo_url]
        if not helpers.is_url(repo_url):
            return gitrepository.GitRepository(repo_url)
        directory = tempfile.TemporaryDirectory(prefix="mknodes_repo_")
        logger.info("Created temporary directory %s", directory.name)
        logger.info("Cloning %s with depth %s", repo_url, clone_depth)
        repo = gitrepository.GitRepository.clone_from(
            repo_url,
            directory.name,
            depth=clone_depth,
        )
        logger.info("Finished cloning.")
        repo.temp_directory = directory
        self._directories[str(repo.working_dir)] = directory
        self._repos[repo_url] = repo
        return repo


registry = RepoRegistry()


if __name__ == "__main__":
    reg = RepoRegistry()
    repo = reg.get_repo("https://github.com/mkdocs/mkdocs")
    print(repo)



from __future__ import annotations

import abc
import os

from typing import Any

from mknodes.pages import mkpage
from mknodes.utils import log


logger = log.get_logger(__name__)


class MkTemplatePage(mkpage.MkPage, metaclass=abc.ABCMeta):
    """MkPage subclass used for rendering templates."""

    def __init__(
        self,
        *args: Any,
        template: str | os.PathLike,
        template_parent: str | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            args: Arguments passed to parent
            template: Template to use
            template_parent: Optional parent template to use
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(*args, **kwargs)
        self.template_main = template
        self.template_parent = template_parent

    @property
    def extra_variables(self) -> dict[str, Any]:
        """Extra variables for the environment. Can be overridden by subclasses."""
        return {}

    def to_markdown(self) -> str:
        env = self.env
        with env.with_globals(**self.extra_variables):
            if isinstance(self.template_main, os.PathLike):
                return env.render_file(self.template_main)
            return env.render_template(
                self.template_main,
                parent_template=self.template_parent,
            )



from __future__ import annotations

from collections.abc import Callable
import dataclasses

from typing import TYPE_CHECKING, Any, Literal


if TYPE_CHECKING:
    import click

    from click import types as clicktypes
    import typer


@dataclasses.dataclass
class Param:
    count: bool = False
    """Whether the parameter increments an integer."""
    default: Any = None
    """The default value of the parameter."""
    envvar: str | None = None
    """ the environment variable name for this parameter."""
    flag_value: bool = False
    """ Value used for this flag if enabled."""
    help: str | None = None  # noqa: A003
    """A formatted help text for this parameter."""
    hidden: bool = False
    """Whether this parameter is hidden."""
    is_flag: bool = False
    """Whether the parameter is a flag."""
    multiple: bool = False
    """Whether the parameter is accepted multiple times and recorded."""
    name: str = ""
    """The name of this parameter."""
    nargs: int = 1
    """The number of arguments this parameter matches."""
    opts: list[str] = dataclasses.field(default_factory=list)
    """Options for this parameter."""
    param_type_name: Literal["option", "parameter", "argument"] = "option"
    """The type of the parameter."""
    prompt: Any = None
    """Whether user is prompted for this parameter."""
    required: bool = False
    """Whether the parameter is required."""
    secondary_opts: list[str] = dataclasses.field(default_factory=list)
    """Secondary options for this parameter."""
    type: clicktypes.ParamType | None = None  # noqa: A003
    """The type object of the parameter."""
    callback: Callable[[click.Context, click.Parameter, Any], Any] | None = None
    """A method to further process the value after type conversion."""
    expose_value: str = ""
    """Whether value is passed onwards to the command callback and stored in context."""
    is_eager: bool = False
    """Whether the param is eager."""
    metavar: str | None = None
    """How value is represented in the help page."""

    @property
    def opt_str(self) -> str:
        """A formatted and sorted string containing the the options."""
        return ", ".join(f"`{i}`" for i in reversed(self.opts))

    def to_markdown(self):
        lines = [f"### {self.opt_str}"]
        if self.required:
            lines.append("**REQUIRED**")
        if self.envvar:
            lines.append(f"**Environment variable:** {self.envvar}")
        if self.multiple:
            lines.append("**Multiple values allowed.**")
        if self.default:
            lines.append(f"**Default:** {self.default}")
        if self.is_flag:
            lines.append(f"**Flag:** {self.flag_value}")
        if self.help:
            lines.append(self.help)
        return "\n\n".join(lines)


@dataclasses.dataclass
class CommandInfo:
    name: str
    """The name of the command."""
    description: str
    """A description for this command."""
    usage: str
    """A formatted string containing a formatted "usage string" (placeholder example)"""
    subcommands: dict[str, CommandInfo] = dataclasses.field(default_factory=dict)
    """A command-name->CommandInfo mapping containing all subcommands."""
    deprecated: bool = False
    """Whether this command is deprecated."""
    epilog: str | None = None
    """Epilog for this command."""
    hidden: bool = False
    """Whether this command is hidden."""
    params: list[Param] = dataclasses.field(default_factory=list)
    """A list of Params for this command."""

    def __getitem__(self, name):
        return self.subcommands[name]

    def to_markdown(self, recursive: bool = False):
        import mknodes as mk

        header = f"## {self.name}\n\n"
        text = header + self.description + "\n\n" + str(mk.MkCode(self.usage))
        params = [i.to_markdown() for i in self.params]
        cmd_text = text + "\n\n\n" + "\n\n\n".join(params)
        if not recursive:
            return cmd_text
        children_text = "\n".join(
            i.to_markdown(recursive=True) for i in self.subcommands.values()
        )
        return cmd_text + children_text


def get_typer_info(
    instance: typer.Typer | click.Group,
    command: str | None = None,
) -> CommandInfo:
    """Return a `CommmandInfo` object for command of given `Typer` instance.

    Arguments:
        instance: A `Typer` or **click** `Group` instance
        command: The command to get info for.
    """
    import typer

    from typer.main import get_command

    cmd = get_command(instance) if isinstance(instance, typer.Typer) else instance
    info = get_command_info(cmd)
    if command:
        ctx = typer.Context(cmd)
        subcommands = getattr(cmd, "commands", {})
        cmds = {k: get_command_info(v, parent=ctx) for k, v in subcommands.items()}
        return cmds.get(command, info)
    return info


def get_command_info(
    command: click.Command,
    parent: click.Context | None = None,
) -> CommandInfo:
    """Get a `CommandInfo` dataclass for given click `Command`.

    Arguments:
        command: The **click** `Command` to get info for.
        parent: The optional parent context
    """
    import typer

    ctx = typer.Context(command, parent=parent)
    subcommands = getattr(command, "commands", {})
    dct = ctx.command.to_info_dict(ctx)
    return CommandInfo(
        name=ctx.command.name or "",
        description=ctx.command.help or ctx.command.short_help or "",
        usage=_make_usage(ctx),
        params=[Param(**i) for i in dct["params"]],
        subcommands={k: get_command_info(v, parent=ctx) for k, v in subcommands.items()},
        deprecated=dct["deprecated"],
        epilog=dct["epilog"],
        hidden=dct["hidden"],
    )


def _make_usage(ctx: click.Context) -> str:
    """Create the full command usage string.

    Arguments:
        ctx: The context to create a usage string for.
    """
    # Gets the usual 'Usage' string without the prefix.
    formatter = ctx.make_formatter()
    pieces = ctx.command.collect_usage_pieces(ctx)
    formatter.write_usage(ctx.command_path, " ".join(pieces), prefix="")
    usage = formatter.getvalue().rstrip("\n")
    # Generate the full usage string based on parents if any, i.e. `root sub1 sub2 ...`.
    full_path = []
    current: click.Context | None = ctx
    while current is not None:
        name = current.command.name.lower() if current.command.name else ""
        full_path.append(name)
        current = current.parent

    full_path.reverse()
    return " ".join(full_path) + usage


if __name__ == "__main__":
    from pprint import pprint

    import mkdocs.__main__

    info = get_typer_info(mkdocs.__main__.cli, command="mkdocs")
    pprint(info.to_markdown(recursive=True))



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, Sequence
import copy

from typing import Self

from mknodes.data import treestyles
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)


class Node:
    """A Node class that can be used to build trees."""

    # __slots__ = ("parent", "children")

    def __init__(self, parent: Self | None = None):
        self._parent = parent
        self.children: list[Self] = []

    def __repr__(self):
        return reprhelpers.get_repr(self)

    def __iter__(self) -> Iterator[Self]:
        yield from self.children

    def __rshift__(self, other: Self):
        """Set children using >> bitshift operator for self >> other.

        Args:
            other (Self): other node, children
        """
        other.parent = self

    def __lshift__(self, other: Self):
        """Set parent using << bitshift operator for self << other.

        Args:
            other (Self): other node, parent
        """
        self._parent = other

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        # if self._parent is not None and self._parent != value:
        #     msg = f"{self!r}: parent {self._parent!r} replaced with {value!r}"
        #     logger.debug(msg)
        self._parent = value

    def __copy__(self) -> Self:
        """Shallow copy self."""
        obj = type(self).__new__(self.__class__)
        obj.__dict__.update(self.__dict__)
        return obj

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result

    def append_child(self, item: Self):
        item.parent = self
        self.children.append(item)

    def insert_children(self, idx: int, items: Sequence[Self]):
        self.children[idx:idx] = items
        for item in items:
            item.parent = self

    @property
    def ancestors(self) -> Iterable[Self]:
        """Get iterator to yield all ancestors of self, does not include self."""
        node = self
        while (node := node.parent) is not None:
            yield node

    @property
    def descendants(self) -> Iterable[Self]:
        """Get iterator to yield all descendants of self, does not include self."""
        yield from preorder_iter(self, filter_condition=lambda _node: _node != self)

    def is_descendant_of(self, kls: type) -> bool:
        """Returns True if any ancestor is of given type."""
        return any(isinstance(i, kls) for i in self.ancestors)

    @property
    def leaves(self) -> Iterable[Self]:
        """Get iterator to yield all leaf nodes from self."""
        yield from preorder_iter(self, filter_condition=lambda _node: _node.is_leaf)

    @property
    def siblings(self) -> Iterable[Self]:
        """Get siblings of self."""
        if self._parent is None:
            return ()
        return tuple(child for child in self._parent.children if child is not self)

    @property
    def left_sibling(self) -> Self | None:
        """Get sibling left of self."""
        if not self._parent:
            return None
        children = self._parent.children
        if child_idx := children.index(self):
            return self._parent.children[child_idx - 1]
        return None

    @property
    def right_sibling(self) -> Self | None:
        """Get sibling right of self."""
        if not self._parent:
            return None
        children = self._parent.children
        child_idx = children.index(self)
        if child_idx + 1 < len(children):
            return self._parent.children[child_idx + 1]
        return None

    @property
    def node_path(self) -> Iterable[Self]:
        """Get tuple of nodes starting from root."""
        if self._parent is None:
            return [self]
        return (*list(self._parent.node_path), self)

    @property
    def is_root(self) -> bool:
        """Get indicator if self is root node."""
        return self._parent is None

    @property
    def is_leaf(self) -> bool:
        """Get indicator if self is leaf node."""
        return not len(list(self.children))

    @property
    def is_first_child(self) -> bool:
        """Get indicator if self is first child of parent."""
        return not bool(self.left_sibling)

    @property
    def is_last_child(self) -> bool:
        """Get indicator if self is last child of parent."""
        return not bool(self.right_sibling)

    @property
    def root(self) -> Self:
        """Get root node of tree."""
        return self if self._parent is None else self._parent.root

    @property
    def depth(self) -> int:
        """Get depth of self, indexing starts from 1."""
        return 1 if self._parent is None else self._parent.depth + 1

    @property
    def max_depth(self) -> int:
        """Get maximum depth from root to leaf node."""
        return max(
            [self.root.depth] + [node.depth for node in list(self.root.descendants)],
        )

    def row(self) -> int:  # sourcery skip: assign-if-exp
        if self._parent:
            return self._parent.children.index(self)  # type: ignore
        return 0

    def pformat(self, indent: int = 0, max_depth: int | None = None):
        lines = [
            _indent * "    " + repr(child_item)
            for _indent, child_item in self.iter_nodes(indent, max_depth)
        ]
        return "\n".join(lines)

    def iter_nodes(self, indent: int = 0, max_depth: int | None = None):
        if max_depth is not None and indent > max_depth:
            return
        yield indent, self
        for child_item in self.children:
            yield from child_item.iter_nodes(indent + 1)

    def get_tree_repr(
        self,
        max_depth: int | None = None,
        style: treestyles.TreeStyleStr | tuple[str, str, str, str] | None = None,
        detailed: bool = True,
    ) -> str:
        def formatter(x) -> str:
            return repr(x) if detailed else x.__class__.__name__

        lines = [
            f"{pre_str}{fill_str}{formatter(_node)}"
            for pre_str, fill_str, _node in self.yield_tree(
                max_depth=max_depth,
                style=style or "ascii",
            )
        ]
        return formatter(self) + "\n" + "\n".join(lines[1:])

    def yield_tree(
        self,
        max_depth: int | None = None,
        style: treestyles.TreeStyleStr | tuple = "const",
    ) -> Iterable[tuple[str, str, Node]]:
        if isinstance(style, tuple):
            custom_style: list[str] = list(style)
            parent_last, filename_middle, filename_last = custom_style
            gap_str = " " * len(parent_last)
        else:
            style_obj = treestyles.STYLES[style]
            parent_last = style_obj.parent_last
            filename_middle = style_obj.filename_middle
            filename_last = style_obj.filename_last
            gap_str = style_obj.parent_middle
        unclosed_depth = set()
        initial_depth = self.depth
        for _node in preorder_iter(self, max_depth=max_depth):
            pre_str = ""
            fill_str = ""
            if not _node.is_root:
                node_depth = _node.depth - initial_depth

                # Get fill_str (filename_middle or filename_last)
                if _node.right_sibling:
                    unclosed_depth.add(node_depth)
                    fill_str = filename_middle
                else:
                    if node_depth in unclosed_depth:
                        unclosed_depth.remove(node_depth)
                    fill_str = filename_last

                pre_str = "".join(
                    parent_last if _depth in unclosed_depth else gap_str
                    for _depth in range(1, node_depth)
                )
            yield pre_str, fill_str, _node

    # def displayable(self, style_name: treestyles.TreeStyleStr = "ascii"):
    #     style = treestyles.STYLES[style_name]
    #     if self.parent is None:
    #         return repr(self)
    #     prefix = style.filename_last if self.is_last_child else style.filename_middle
    #     parts = [f"{prefix!s} {self!r}"]
    #     parent = self.parent
    #     while parent and parent.parent is not None:
    #         part = style.parent_last if parent.is_last_child else style.parent_middle
    #         parts.append(part)
    #         parent = parent.parent
    #     return "".join(reversed(parts))

    # def get_tree_repr(self, style: treestyles.TreeStyleStr = "ascii"):
    #     nodes = [self, *list(self.descendants)]
    #     return "\n".join(i.displayable(style) for i in nodes)


def preorder_iter(
    tree: Node,
    filter_condition: Callable[[Node], bool] | None = None,
    stop_condition: Callable[[Node], bool] | None = None,
    max_depth: int | None = None,
) -> Iterable[Node]:
    """Iterate through all children of a tree.

    It is topologically sorted because a parent node is processed before its child nodes.

    a
    ├── b
    │   ├── d
    │   └── e
    │       ├── g
    │       └── h
    └── c
        └── f

    >>> [node.node_name for node in preorder_iter(root)]
    ['a', 'b', 'd', 'e', 'g', 'h', 'c', 'f']

    Arguments:
        tree: input tree
        filter_condition: function that takes in node as argument
        stop_condition: function that takes in node as argument
        max_depth: maximum depth of iteration, based on `depth` attribute
    """
    if (not max_depth or tree.depth <= max_depth) and (
        not stop_condition or not stop_condition(tree)
    ):
        if not filter_condition or filter_condition(tree):
            yield tree
        for child in tree.children:
            yield from preorder_iter(child, filter_condition, stop_condition, max_depth)


if __name__ == "__main__":
    node = Node()
    sub = Node(parent=node)
    subsub = Node(parent=sub)
    node.children = [sub, subsub]



from __future__ import annotations

from collections.abc import Sequence

from mknodes.basenodes import mkcontainer, mknode
from mknodes.info import linkprovider
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


class MkList(mkcontainer.MkContainer):
    """Node for showing a formatted list."""

    ICON = "octicons/list-unordered-24"
    REQUIRED_EXTENSIONS = [resources.Extension("def_list")]

    def __init__(
        self,
        items: Sequence[str | mknode.MkNode] | None = None,
        *,
        ordered: bool = False,
        shorten_after: int | None = None,
        as_links: bool = False,
        header: str = "",
    ):
        """Constructor.

        Arguments:
            items: Items to show as a list
            ordered: whether the list should be numbered
            shorten_after: will clip the list after n items and append a "..."
            as_links: will convert the label to links which will get replaced by plugin
            header: Section header
        """
        # if as_links:
        #     items = [link.Link(i) for i in items]
        items = items or []
        self.ordered = ordered
        self.shorten_after = shorten_after
        self.as_links = as_links
        super().__init__(content=list(items), header=header)

    def __len__(self):
        return len(self.items)

    def __repr__(self):
        items = [reprhelpers.to_str_if_textnode(i) for i in self.items]
        return reprhelpers.get_repr(
            self,
            items=items,
            shorten_after=self.shorten_after,
            as_links=self.as_links,
            _filter_empty=True,
            _filter_false=True,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        list_1 = MkList(["Item 1", "Item 2", "Item 3"])
        # list can also have a max length. they will get shortened with a "..." entry.
        list_2 = MkList(["Item"] * 6, shorten_after=3)
        # They can also be ordered.
        list_3 = MkList(["Item 1", "Item 2", "Item 3"], ordered=True)
        # and can contain markdown.
        list_4 = MkList([mk.MkAdmonition("Markup")] * 3)

        page += mk.MkReprRawRendered(list_1, header="### Regular")
        page += mk.MkReprRawRendered(list_2, header="### Shortened")
        page += mk.MkReprRawRendered(list_3, header="### Ordered")
        page += mk.MkReprRawRendered(list_4, header="### Containing markdown")

    def _prep(self, item) -> str:
        return linkprovider.linked(str(item)) if self.as_links else str(item)

    def _to_markdown(self) -> str:
        if not self.items:
            return ""
        prefix = "1." if self.ordered else "*"
        lines = [f"  {prefix} {self._prep(i)}" for i in self.items[: self.shorten_after]]
        if self.shorten_after and len(self.items) > self.shorten_after:
            lines.append(f"  {prefix} ...")
        return "\n".join(lines) + "\n"

    def to_html(self) -> str:
        """Formats list in html as one single line.

        Can be useful for including in Tables.
        """
        if not self.items:
            return ""
        tag_name = "ol" if self.ordered else "ul"
        items = [f"<li>{self._prep(i)}</li>" for i in self.items[: self.shorten_after]]
        item_str = "".join(items)
        if self.shorten_after and len(self.items) > self.shorten_after:
            item_str += "<li>...</li>"
        return f"<{tag_name}>{item_str}</{tag_name}>"


if __name__ == "__main__":
    section = MkList(["a", "b"], header="test")
    print(section.to_html())



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from mknodes.basenodes import mkbinaryimage, mknode
from mknodes.utils import log, reprhelpers, resources, xmlhelpers as xml


if TYPE_CHECKING:
    from mknodes.navs import mknav
    from mknodes.pages import mkpage


logger = log.get_logger(__name__)

CARD_DEFAULT_SIZE = 200


def build_html_card(
    *,
    image: str,
    title: str,
    link: str | None = None,
    size: int = CARD_DEFAULT_SIZE,
    caption: str | None = None,
    path_dark_mode: str | None = None,
) -> xml.HTMLElement:
    root = xml.A(href=link) if link else xml.Div()
    card_div = xml.Div("card", parent=root)
    container_div = xml.Div("showcase-container", parent=card_div)
    src = f"{image}#only-light" if path_dark_mode else image
    style = f"width:{size}px,height:{size}px"
    xml.Img(parent=container_div, src=src, alt=title, style=style)
    if path_dark_mode:
        src = f"{path_dark_mode}#only-dark"
        style = f"width:{size}px,height:{size}px"
        xml.Img(parent=container_div, src=src, alt=title, style=style)
    if caption:
        xml.Div("overlay", text=caption, parent=container_div)
    p = xml.P(parent=card_div)
    button = xml.Button(parent=p)
    button.text = title
    return root


class MkCard(mknode.MkNode):
    """A card node, displaying an image, a button-like label and a hover caption.

    This node requires addtional CSS to work.
    """

    ICON = "material/square-medium"
    STATUS = "new"
    CSS = [resources.CSSFile("css/grid.css")]

    def __init__(
        self,
        title: str,
        image: str,
        caption: str | None = None,
        target: str | mkpage.MkPage | mknav.MkNav | None = None,
        size: int = CARD_DEFAULT_SIZE,
        path_dark_mode: str | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            title: Button text
            image: Card image
            caption: Image caption
            target: Link target. Can be a URL, an MkPage, or an MkNav
            size: Height/Width of the card
            path_dark_mode: Optional alternative image for dark mode
            kwargs: keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.target = target
        self.title = title
        self.image = image
        self.caption = caption
        self.size = size
        self.path_dark_mode = path_dark_mode

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            title=self.title,
            image=self.image,
            caption=self.caption,
            target=self.target,
            size=self.size if self.size != CARD_DEFAULT_SIZE else None,
            path_dark_mode=self.path_dark_mode,
            _filter_empty=True,
        )

    @property
    def url(self) -> str:
        return self.ctx.links.get_url(self.target) if self.target else ""

    def _to_markdown(self) -> str:
        elem = build_html_card(
            title=self.title,
            image=self.image,
            caption=self.caption,
            link=self.url,
            size=self.size,
            path_dark_mode=self.path_dark_mode,
        )
        return elem.to_string()

    @classmethod
    def for_page(cls, page):
        image = mkbinaryimage.MkBinaryImage.for_icon(page.icon)
        card = MkCard(
            page.title,
            image=image.path,
            caption=page.subtitle or "",
            target=page,
        )
        card.add_file(image.path, image.data)
        return card

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        img = "https://picsum.photos/300"
        url = "https://phil65.github.io/mknodes/"
        img_dark = "https://picsum.photos/200"
        node = MkCard(image=img, title="Title", target=url)
        page += mk.MkReprRawRendered(node, header="### Without caption")

        node = MkCard(image=img, title="Title", target=url, caption="Caption")
        page += mk.MkReprRawRendered(node, header="### With caption")

        node = MkCard(image=img, title="Title", path_dark_mode=img_dark)
        page += mk.MkReprRawRendered(node, header="### Separate dark mode image")


if __name__ == "__main__":
    from mknodes.pages import mkpage

    page = mkpage.MkPage("test", icon="material/puzzle-edit")
    card = MkCard.for_page(page)
    print(card)



"""Filesystem paths."""

from __future__ import annotations

import pathlib


ROOT = pathlib.Path(__file__).parent

TEST_RESOURCES = ROOT.parent / "tests" / "data"
RESOURCES = ROOT / "resources"

DEFAULT_BUILD_FN = "mknodes:MkDefaultWebsite.for_project"



from __future__ import annotations

import mknodes as mk

from mknodes.info import contexts


NAMESPACES = {
    "`metadata`": "Package information",
    "`git`": "Local repository information",
    "`github`": "Information about the remote repository",
    "`theme`": "Information about the theme being used",
}

CONTEXTS = [
    contexts.PackageContext,
    contexts.ThemeContext,
    contexts.GitContext,
    contexts.GitHubContext,
]

nav = mk.MkNav("Templating")


@nav.route.page(is_index=True)
def _(page: mk.MkPage):
    page += mk.MkJinjaTemplate("templating/template_index.jinja")


@nav.route.nav("Jinja Namespace")
def _(nav: mk.MkNav):
    def add_context_doc(container, context):
        container += mk.MkDocStrings(
            context,
            show_root_toc_entry=False,
            show_if_no_docstring=True,
            heading_level=4,
            show_bases=False,
            show_source=False,
        )

    for ctx in CONTEXTS:
        subpage = nav.add_page(ctx.__name__)
        add_context_doc(subpage, ctx)
    page = nav.add_page(is_index=True, hide="toc")
    variables = dict(namespaces=NAMESPACES)
    template = "templating/template_namespace_index.jinja"
    page += mk.MkJinjaTemplate(template, variables=variables)


@nav.route.page("Utility filters")
def _(page: mk.MkPage):
    page += mk.MkJinjaTemplate("templating/template_filters.jinja")


@nav.route.page("MkNode filters")
def _(page: mk.MkPage):
    page += mk.MkJinjaTemplate("templating/jinja_mknode_filters.jinja")



from __future__ import annotations

from mknodes.basenodes import mkbasetable
from mknodes.utils import log, xmlhelpers as xml


logger = log.get_logger(__name__)


class MkHtmlTable(mkbasetable.MkBaseTable):
    """Class representing a html table.

    Compared to MkTable, this will end up with a more verbose output,
    but it can contain more complex Markdown in cells.
    """

    STATUS = "new"

    def get_element(self) -> xml.Table | None:
        table_data = self.data  # property
        root = xml.Table(markdown=True)
        if not any(table_data[k] for k in table_data):
            return None
        data = [[str(k) for k in row] for row in self.iter_rows()]
        headers = list(table_data.keys())
        data.insert(0, headers)
        for items in data:
            tr = xml.Tr(parent=root)
            for item in items:
                td = xml.Td(parent=tr)
                td.text = "\n" + item + "\n"
        return root

    def _to_markdown(self) -> str:
        root = self.get_element()
        return root.to_string(space="") if root is not None else ""

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        code_col = [mk.MkCode("print('hello world')\nsys.exit()") for _ in range(3)]
        admonitions = [mk.MkAdmonition("Admonition inside cell") for _ in range(3)]
        tabs = [mk.MkTabbed(dict(A=["Tab a"], B=["Tab b"])) for _ in range(3)]
        data: dict[str, list] = dict(Code=code_col, Admonitions=admonitions, Tabs=tabs)
        page += mk.MkReprRawRendered(MkHtmlTable(data))


if __name__ == "__main__":
    table = MkHtmlTable(data={"Column A": ["A", "B", "C"], "Column B": ["C", "D", "E"]})
    print(table._to_markdown())



from __future__ import annotations

from typing import Any

from mknodes.basenodes import mkcode, mkcontainer, mkheader, mknode, mktext
from mknodes.data import installmethods
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)


class MkInstallGuide(mkcontainer.MkContainer):
    """Node to display an install guide."""

    ICON = "material/help"

    def __init__(
        self,
        distribution: str | None = None,
        package_repos: list[installmethods.InstallMethodStr] | None = None,
        header_level: int = 3,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            distribution: name of the distribution to install
            package_repos: package repositories the project is available on
            header_level: Header level for each section
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self._distribution = distribution
        self.header_level = header_level
        self._package_repos = package_repos

    @property
    def package_repos(self) -> list[installmethods.InstallMethodStr]:
        if self._package_repos:
            return self._package_repos
        return self.ctx.metadata.package_repos or ["pip"]

    @property
    def distribution(self):
        return self._distribution or self.ctx.metadata.distribution_name

    @distribution.setter
    def distribution(self, value):
        self._distribution = value

    @property
    def items(self) -> list[mknode.MkNode]:
        if not self.distribution:
            return []
        klasses = [installmethods.InstallMethod.by_id(i) for i in self.package_repos]
        methods = [i(self.distribution) for i in klasses]
        return [self.get_section_for(method) for method in methods]

    @items.setter
    def items(self, value):
        pass

    def get_section_for(
        self,
        method: installmethods.InstallMethod,
    ) -> mkcontainer.MkContainer:
        items = [
            mkheader.MkHeader(method.ID, level=self.header_level),
            mktext.MkText(method.info_text()),
            mkcode.MkCode(method.install_instructions()),
        ]
        # proj = self.associated_distribution
        # if method.ID == "pip" and proj and (extras := proj.info.extras):
        #     extras_str = ",".join(extras)
        #     text = f"{method.install_instructions()}[{extras_str}]"
        #     code = mkcode.MkCode(text)
        #     items.append(code)
        return mkcontainer.MkContainer(items, parent=self)

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            distribution=self._distribution,
            package_repos=self._package_repos,
            header_level=self.header_level,
            _filter_empty=True,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        # this will show an install guide for our associated project
        node = MkInstallGuide()
        page += mk.MkReprRawRendered(node, header="### From project")

        # we can also explicitely define the repositories
        node = MkInstallGuide("mkdocs", package_repos=["pipx"])
        page += mk.MkReprRawRendered(node, header="### Explicit")


if __name__ == "__main__":
    guide = MkInstallGuide.with_default_context()
    print(guide)



from __future__ import annotations

from typing import Literal

from mknodes.basenodes import mkcontainer, mknode
from mknodes.utils import log, reprhelpers, resources, xmlhelpers as xml


logger = log.get_logger(__name__)


class MkSpeechBubble(mkcontainer.MkContainer):
    """Node for showing a css-based speech bubble."""

    ICON = "material/chat"
    CSS = [resources.CSSFile("css/speechbubble.css")]

    def __init__(
        self,
        content: str | mknode.MkNode | list | None = None,
        arrow: Literal["top", "bottom", "left", "right"] | None = "bottom",
        **kwargs,
    ):
        self.arrow = arrow
        super().__init__(content=content or [], **kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(self, content=self.items, arrow=self.arrow)

    def get_element(self) -> xml.Div:
        klass = f"speech {self.arrow}" if self.arrow else "speech"
        root = xml.Div(klass, markdown=True)
        root.text = super()._to_markdown()
        return root

    def _to_markdown(self) -> str:
        return self.get_element().to_string()

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkSpeechBubble(MkSpeechBubble.__doc__)
        page += mk.MkReprRawRendered(node)
        node = MkSpeechBubble(MkSpeechBubble.__doc__, arrow="left")
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    grid = MkSpeechBubble("test")
    print(grid)



import mknodes as mk


def a_quick_tour(page: mk.MkPage):
    # This will be a quick, short random introduction of some of the nodes
    # included in **MkNodes**.
    # The selection of nodes is totally random, this should just provide a quick overview
    # how to interact with the nodes.
    #
    # Let the tour begin!
    #
    # Our first [MkNode][mknodes.MkNode] is very clever. You just show him a node and he
    # will tell you all you need to know about him.
    # It's [MkDocStrings][mknodes.MkDocStrings]!
    # Lets check what he knows about [MkPage][mknodes.MkPage]:
    node = mk.MkDocStrings(mk.MkPage)
    # [MkDocStrings][mknodes.MkDocStrings] sometimes really writes long stories,
    # so we will put everything into
    # a collapsed [MkAdmonition][mknodes.MkAdmonition] box,
    # so we dont need that much space:
    admonition = mk.MkAdmonition(node, collapsible=True)
    str(admonition)
    # Here is the result:
    # {{ "mknodes.MkPage" | MkDocStrings | MkAdmonition(collapsible=True) }}

    # [MkDocStrings][mknodes.MkDocStrings] can even show information about himself.
    # Very talented!
    node = mk.MkDocStrings(mk.MkDocStrings)
    admonition = mk.MkAdmonition(node, collapsible=True)
    str(admonition)
    # {{ "mknodes.MkDocStrings" | MkDocStrings | MkAdmonition(collapsible=True)}}

    # Okay, that's enough DocStrings for today. You can find more of
    # [MkDocStrings][mknodes.MkDocStrings] work
    # in the API documentation, he will tell you something about every
    # [MkNode][mknodes.MkNode] there.
    #
    # Another [MkNode][mknodes.MkNode] who is displaying his skills in the API docs is
    # [MkClassDiagram][mknodes.MkClassDiagram].
    #
    # He's a very talented painter. Perhaps he can draw us something about
    # [MkPage][mknodes.MkPage]!
    #
    diagram = mk.MkClassDiagram(mk.MkPage)
    str(diagram)
    # {{ "mknodes.MkPage" | MkClassDiagram }}

    # [MkClassDiagram][mknodes.MkClassDiagram] can draw different kind of graphs.
    # The first picture [MkClassDiagram][mknodes.MkClassDiagram] has painted was about
    # base classes. Lets check out the subclasses:
    diagram = mk.MkClassDiagram(mk.MkPage, mode="subclasses")
    str(diagram)
    # {{ "mknodes.MkPage" | MkClassDiagram(mode="subclasses") }}

    # There are multiple talented drawers among the **MkNodes**.
    # [MkPipDepTree][mknodes.MkPipDepTree] is known for his dependency drawings,
    # we can ask him to draw a graph for one of our dependencies.
    #
    # To not overboard him, lets pick a package without too many dependencies:

    node = mk.MkPipDepTree("gitpython", direction="LR")
    str(node)
    # {{ "gitpython" | MkPipDepTree(direction="LR") }}

    # We now come to the last [MkNode][mknodes.MkNode] of our quick tour.
    #
    # Let's introduce [MkMetadataBadges][mknodes.MkMetadataBadges]!
    #
    # [MkMetadataBadges][mknodes.MkMetadataBadges] just loves Badges. He creates them
    # himself and doesnt rely on webservies.
    node = mk.MkMetadataBadges("classifiers", package="mkdocstrings")
    str(node)
    # {{ "classifiers" | MkMetadataBadges(package="mkdocstrings") }}

    # Looks neat, right?
    # This is also a good chance to explain how nodes can get their information
    # from the context. Let me explain:
    #
    # If we instanciate **MkNodes** and dont add them to the tree,
    # then the nodes are clueless. They dont know who they belong to.
    #
    # To make them part of the tree, we either add them for example to a page
    #
    # (like this: `page += node`)
    #
    # or we pass them a parent on instantiation.
    #
    # In this example, once we add [MkMetadataBadges][mknodes.MkMetadataBadges]
    # to the tree, that node knows
    # that he should create badges for our very own package, **mknodes**,
    # unless we explicitely tell him to do otherwise. It will become his new "default".
    #
    # You dont believe me? Let me show you:
    node = mk.MkMetadataBadges("websites", parent=page)

    # Now that node is connected. If we ask him to draw now, he will create website
    # badges for **mknodes**!

    str(node)
    # {{ "websites" | MkMetadataBadges }}

    # This mechanism is the same for many **MkNodes**. For example, our earlier guest
    # [MkPipDepTree][mknodes.MkPipDepTree] behaves the same. Once connected,
    # he will want to paint **mknodes** dependency graphs without us telling him to do so!
    #
    # Interesting, right?
    #
    # That's it for a first quick look at the nodes.
    # There are about 70 different ones available in this package.
    # Some are exciting, some are boring. It's no difference to humans.
    #
    # Oh, and if you wonder how this tour was done: I got help from
    # [MkCommentedCode][mknodes.MkCommentedCode]!
    #
    page += mk.MkCommentedCode(a_quick_tour, style="text")

    # [MkCommentedCode][mknodes.MkCommentedCode] parses a function and separates comments
    # from code. These chunks are displayed on a rotating basis then.
    #
    # Before we end the tour, let's take a look at the raw material.
    # We can use the [MkCode][mknodes.MkCode] node for that.
    #
    node = mk.MkCode.for_object(a_quick_tour)
    text = str(node).replace(r"{", "<").replace(r"}", ">")
    page += text
    # As you can see, we added the [MkCode][mknodes.MkCode] node to the page.
    # It will be displayed right after the [MkCommentedCode][mknodes.MkCommentedCode]
    # block. You will see it right below.
    #
    # Oh, and the String replacement you probably noticed is a quick hack to prevent
    # **jinja2** code from getting executed. YOu can ignore that, it's not relevant
    #  for this tour.
    #
    # That's it. The rest of the nodes you need to check out yourself. Have fun!



import mknodes as mk


nav = mk.MkNav("Development")


@nav.route.page("Module overview", icon="file-tree-outline")
def _(page: mk.MkPage):
    page += mk.MkModuleOverview(maximum_depth=2)


@nav.route.page("Code of conduct", icon="octicons/code-of-conduct-24")
def _(page: mk.MkPage):
    page += mk.MkCodeOfConduct()


@nav.route.page("Contributing", icon="help")
def _(page: mk.MkPage):
    page += mk.MkCommitConventions()
    page += mk.MkPullRequestGuidelines()


@nav.route.page("License", hide="toc", icon="license")
def _(page: mk.MkPage):
    page += mk.MkLicense()


@nav.route.page("Dependencies", hide="toc", icon="database")
def _(page: mk.MkPage):
    page += mk.MkDependencyTable(layout="badge")
    page += mk.MkPipDepTree(direction="LR")


@nav.route.page("Development environment", icon="dev-to")
def _(page: mk.MkPage):
    page += mk.MkDevEnvSetup()


@nav.route.page("Dev Tools", icon="wrench")
def _(page: mk.MkPage):
    page += mk.MkDevTools()



from __future__ import annotations

import collections

from collections.abc import ItemsView, KeysView, ValuesView
import pprint
from types import MappingProxyType, SimpleNamespace
from typing import Any, ClassVar, Protocol, runtime_checkable

from mknodes.basenodes import mkcode
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)


@runtime_checkable
class IsDataclass(Protocol):
    __dataclass_fields__: ClassVar[dict]


class MkPrettyPrint(mkcode.MkCode):
    """Node to show a prettyprinted data structure."""

    ICON = "material/printer"

    def __init__(
        self,
        obj: (
            dict
            | list
            | str
            | tuple
            | set
            | bytes
            | bytearray
            | MappingProxyType
            | SimpleNamespace
            | ValuesView
            | KeysView
            | collections.Counter
            | collections.ChainMap
            | collections.deque
            | collections.UserDict
            | collections.UserList
            | collections.UserString
            | ItemsView
            | IsDataclass
        ),
        *,
        indent: int = 1,
        depth: int | None = None,
        width: int = 80,
        compact: bool = False,
        sort_dicts: bool = False,
        underscore_numbers: bool = False,
        header: str = "",
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            obj: Object to prettyprint
            header: Section header
            indent: Specifies the amount of indentation added for each nesting level
            depth: Maximum nesting depth to print
            width: Specifies the desired maximum number of characters per line
            compact: Compact format for long sequences
            sort_dicts: Whether dicts should be sorted after keys
            underscore_numbers: Whether to use underscore as a separator for long numbers
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(header, **kwargs)
        self.obj = obj
        self.print_indent = indent  # avoid name conflicts
        self.print_width = width
        self.print_depth = depth
        self.print_compact = compact
        self.print_sort_dicts = sort_dicts
        self.print_underscore_numbers = underscore_numbers

    @property
    def text(self):
        return pprint.pformat(
            self.obj,
            indent=self.print_indent,
            width=self.print_width,
            depth=self.print_depth,
            compact=self.print_compact,
            sort_dicts=self.print_sort_dicts,
            underscore_numbers=self.print_underscore_numbers,
        )

    @text.setter
    def text(self, text):
        self.obj = text

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            obj=self.obj,
            indent=self.print_indent,
            width=self.print_width,
            depth=self.print_depth,
            compact=self.print_compact,
            sort_dicts=self.print_sort_dicts,
            underscore_numbers=self.print_underscore_numbers,
            _filter_empty=True,
            _filter_false=True,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkPrettyPrint(obj=[dict(a="test " * 5)] * 5)
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    section = MkPrettyPrint([dict(a="test " * 5)] * 5, header="test")
    print(section.to_markdown())



from __future__ import annotations

from collections.abc import Sequence
import functools

from typing import Any, Literal
import xml.etree.ElementTree as etree

import markdown

from mknodes import paths


PYCONIFY_TO_PREFIXES = {
    "mdi": "material",
    "simple-icons": "simple",
    "octicon": "octicons",
    "fa6-regular": "fontawesome-regular",
    "fa-brands": "fontawesome-brands",
    "fa6-solid": "fontawesome-solid",
}

Rotation = Literal["90", "180", "270", 90, 180, 270, "-90", 1, 2, 3]
Flip = Literal["horizontal", "vertical", "horizontal,vertical"]


def icon_for_url(url: str) -> str | None:
    """Return a pyconify icon key for given url."""
    from urllib import parse

    socials = {
        "matrix.to": "fa-brands:gitter",
        "x.com": "fa-brands:twitter",
        "fosstodon.org": "fa-brands:mastodon",
    }
    netloc = parse.urlsplit(url).netloc.lower()
    stem = netloc.split(".")[-2]
    for k, v in socials.items():
        if k == netloc:
            return v
    idx = _get_pyconify_icon_index()
    if (name := f":fa-brands-{stem}:") in idx:
        return get_pyconify_key(idx[name]["name"])
    return None


@functools.cache
def _get_collection_map(*prefixes: str) -> dict[str, list[str]]:
    """Return a dictionary with a mapping from pyconify name to icon prefixes.

    In order to provide compatibility with the materialx-icon-index,
    we also add the prefixes used by that index, which is different from
    the pyconify prefixes. (material vs mdi etc, see PYCONIFY_TO_PREFIXES)
    """
    import pyconify

    mapping = {coll: [coll] for coll in pyconify.collections(*prefixes)}
    for k, v in PYCONIFY_TO_PREFIXES.items():
        if k in mapping:
            mapping[k].append(v)
    return mapping


@functools.cache
def _get_pyconify_icon_index(*collections: str) -> dict[str, dict[str, str]]:
    """Return a icon index for the pymdownx emoji extension containing pyconify icons.

    The dictionaries contain two key-value pairs:
    "name" is the emoji identifier,
    "path" is the pyconify key

    """
    import pyconify

    index = {}
    for coll, prefixes in _get_collection_map(*collections).items():
        collection = pyconify.collection(coll)
        for icon_name in collection.get("uncategorized", []):
            for prefix in prefixes:
                name = f":{prefix}-{icon_name}:"
                index[name] = {
                    "name": name,
                    "path": f"{coll}:{icon_name}",
                    "set": coll,
                }
        for cat in pyconify.collection(coll).get("categories", {}).values():
            for icon_name in cat:
                for prefix in prefixes:
                    name = f":{prefix}-{icon_name}:"
                    index[name] = {
                        "name": name,
                        "path": f"{coll}:{icon_name}",
                        "set": coll,
                    }
    return index


@functools.cache
def _patch_index_with_sets(icon_sets: Sequence[str]) -> dict[str, Any]:
    from pymdownx import twemoji_db

    # Copy the Twemoji index
    index = {
        "name": "twemoji",
        "emoji": twemoji_db.emoji,
        "aliases": twemoji_db.aliases,
    }
    # icons = _get_pyconify_icon_index(*icon_sets)
    icons = load_icon_index()
    index["emoji"].update(icons)
    return index


def twemoji(options: dict[str, Any], md) -> dict[str, Any]:
    """Provide a copied Twemoji index with additional codes for Pyconify icons.

    Used for pymdownx.emoji.
    """
    default = list(PYCONIFY_TO_PREFIXES.keys())
    icon_sets = options.get("icon_sets", default)[:]
    return _patch_index_with_sets(tuple(icon_sets))


def to_svg(
    index: str,
    shortname: str,
    alias: str,
    uc: str,
    alt: str,
    title: str,
    category: str,
    options: dict[str, str],
    md: markdown.Markdown,
):
    """Return svg element (wrapped in a span element).

    Used for pymdownx.emoji.
    """
    from pymdownx.emoji import TWEMOJI_SVG_CDN, add_attributes

    is_unicode = uc is not None
    if is_unicode:
        image_path = options.get("image_path", TWEMOJI_SVG_CDN)
        attributes = {
            "class": options.get("classes", index),
            "alt": alt,
            "src": f"{image_path}{uc}.svg",
        }

        if title:
            attributes["title"] = title

        add_attributes(options, attributes)

        return etree.Element("img", attributes)
    el = etree.Element("span", {"class": options.get("classes", index)})
    svg_path = md.inlinePatterns["emoji"].emoji_index["emoji"][shortname]["path"]
    svg = get_icon_svg(svg_path)
    el.text = md.htmlStash.store(svg)
    return el


def get_icon_svg(
    icon: str,
    color: str | None = None,
    height: str | int | None = None,
    width: str | int | None = None,
    flip: Flip | None = None,
    rotate: Rotation | None = None,
    box: bool | None = None,
) -> str:
    """Return svg for given pyconify icon key.

    Key should look like "mdi:file"
    For compatibility, this method also supports compatibility for
    emoji-slugs (":material-file:") as well as material-paths ("material/file")

    If no icon group is supplied as part of the string, mdi is assumed as group.

    When passing a string with "|" delimiters, the returned string will contain multiple
    icons.

    Arguments:
        icon: Pyconify icon name
        color: Icon color. Replaces currentColor with specific color, resulting in icon
               with hardcoded palette.
        height: Icon height. If only one dimension is specified, such as height, other
                dimension will be automatically set to match it.
        width: Icon width. If only one dimension is specified, such as height, other
               dimension will be automatically set to match it.
        flip: Flip icon.
        rotate: Rotate icon. If an integer is provided, it is assumed to be in degrees.
        box: Adds an empty rectangle to SVG that matches the icon's viewBox. It is needed
            when importing SVG to various UI design tools that ignore viewBox. Those
            tools, such as Sketch, create layer groups that automatically resize to fit
            content. Icons usually have empty pixels around icon, so such software crops
            those empty pixels and icon's group ends up being smaller than actual icon,
            making it harder to align it in design.

    Example:
        get_icon_svg("file")  # implicit mdi group
        get_icon_svg("mdi:file")  # pyconify key
        get_icon_svg("material/file")  # Material-style path
        get_icon_svg(":material-file:")  # material-style emoji slug
        get_icon_svg("mdi:file|:material-file:")  # returns a string with two svgs
    """
    label = ""
    for splitted in icon.split("|"):
        key = get_pyconify_key(splitted)
        import pyconify

        label += pyconify.svg(
            key,
            color=color,
            height=height,
            width=width,
            flip=flip,
            rotate=rotate,
            box=box,
        ).decode()
    return label


def get_pyconify_key(icon: str) -> str:
    """Convert given string to a pyconify key.

    Converts the keys from MkDocs-Material ("material/sth" or ":material-sth:")
    to their pyconify equivalent.

    Arguments:
        icon: The string which should be converted to a pyconify key.
    """
    for k, v in PYCONIFY_TO_PREFIXES.items():
        path = f'{v.replace("-", "/")}/'
        icon = icon.replace(path, f"{k}:")
        icon = icon.replace(f":{v}-", f"{k}:")
    icon = icon.strip(":")
    mapping = {k: v[0] for k, v in _get_collection_map().items()}
    for prefix in mapping:
        if icon.startswith(f"{prefix}-"):
            icon = icon.replace(f"{prefix}-", f"{prefix}:")
            break
    if (count := icon.count(":")) > 1:
        icon = icon.replace(":", "-", count - 1)
    if ":" not in icon:
        icon = f"mdi:{icon}"
    return icon


def get_emoji_slug(icon: str) -> str:
    """Return a icon string which can be used in markdown texts.

    The icon string will get picked up by pymdownx.emoji extension.

    Arguments:
        icon: The string to convert to an emoji slug.
    """
    return f":{get_pyconify_key(icon).replace(':', '-')}:"


def get_icon_xml(icon: str) -> etree.Element:
    """Return a xmlElement for given MaterialIcon.

    Arguments:
        icon: Icon to fetch. If iconname is not explicit (aka does not contain "/"),
              then it will try to get the icon from material/ folder.
    """
    etree.register_namespace("", "http://www.w3.org/2000/svg")
    svg_text = get_icon_svg(icon)
    return etree.fromstring(svg_text)


def write_icon_index():
    """Fetch the complete icon index and write it gzipped to disk."""
    import gzip
    import json

    mapping = _get_pyconify_icon_index()
    path = paths.RESOURCES / "icons.json.gzip"
    with gzip.open(path, "w") as file:
        file.write(json.dumps(mapping).encode())


def load_icon_index() -> dict:
    """Load the complete icon index from disk."""
    import gzip
    import json

    path = paths.RESOURCES / "icons.json.gzip"
    with gzip.open(path, "r") as file:
        return json.loads(file.read())


if __name__ == "__main__":
    idx = load_icon_index()
    print(idx)



from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Literal

import markdown

from mknodes.utils import log


logger = log.get_logger(__name__)


class MdConverter(markdown.Markdown):
    def __init__(
        self,
        extensions: Sequence[str | markdown.Extension] | None = None,
        extension_configs: dict[str, dict[str, Any]] | None = None,
        output_format: Literal["xhtml", "html"] = "html",
        tab_length: int = 4,
    ):
        exts: Sequence[str | markdown.Extension] = [
            "pymdownx.emoji",
            "md_in_html",
            "attr_list",
        ]
        if extensions:
            exts = list({*exts, *extensions})
        super().__init__(
            extensions=exts,
            extension_configs=extension_configs or {},
            output_format=output_format,
            tab_length=tab_length,
        )

    def register_extensions(self, extensions: dict):
        ext_names = list(extensions.keys())
        self.registerExtensions(ext_names, extensions)



from __future__ import annotations

import dataclasses

from typing import Literal


@dataclasses.dataclass
class TaskRunner:
    identifier: TaskRunnerStr
    website: str
    filenames: list[str]
    help_cmd: list[str]


TaskRunnerStr = Literal["makefile", "task", "just", "duty", "invoke", "doit"]


# class MakeFile(TaskRunner):
#     def get_commands(self) -> dict[str, str]:
#         import re

#         dct = {}
#         text = ""
#         for line in text.splitlines():
#             if match := re.match(r"^([a-zA-Z_-]+):.*?## (.*)$$", line):
#                 target, help_text = match.groups()
#                 dct[target] = help_text
#         return dct


makefile = TaskRunner(
    identifier="makefile",
    website="https://www.gnu.org/software/make/manual/make.html",
    filenames=["Makefile"],
    help_cmd=["make", "help"],
)

task = TaskRunner(
    identifier="task",
    website="https://taskfile.dev/",
    filenames=[
        "Taskfile.yml",
        "taskfile.yml",
        "Taskfile.yaml",
        "taskfile.yaml",
        "Taskfile.dist.yml",
        "taskfile.dist.yml",
        "Taskfile.dist.yaml",
        "taskfile.dist.yaml",
    ],
    help_cmd=["task", "--list"],
)

just = TaskRunner(
    identifier="just",
    website="https://github.com/casey/just",
    filenames=["justfile"],
    help_cmd=["just", "--list"],
)

duty = TaskRunner(
    identifier="duty",
    website="https://github.com/pawamoy/duty",
    filenames=["duties.py"],
    help_cmd=["duty", "--list"],
)

invoke = TaskRunner(
    identifier="invoke",
    website="https://www.pyinvoke.org",
    filenames=["invoke.yaml", "invoke.yml", "invoke.json", "invoke.py"],
    help_cmd=["invoke", "--list"],
)

doit = TaskRunner(
    identifier="doit",
    website="http://pydoit.org",
    filenames=["dodo.py"],
    help_cmd=["doit", "list"],
)

TASK_RUNNERS: dict[TaskRunnerStr, TaskRunner] = {
    p.identifier: p for p in [makefile, task, just, duty, invoke, doit]
}



from __future__ import annotations

from collections.abc import Sequence
import os

from typing import Any

from mknodes.basenodes import mknode
from mknodes.utils import helpers, log, reprhelpers, resources


logger = log.get_logger(__name__)


TEXT = """<div data-terminal>
  <span data-ty="input">{input}</span>
  <span data-ty>{output}</span>
</div>"""


class MkCommandOutput(mknode.MkNode):
    """Node to display the terminal output of a command."""

    ICON = "material/bash"
    STATUS = "new"
    CSS = [resources.CSSFile("css/terminal.css")]

    def __init__(self, call: Sequence[str], **kwargs: Any):
        """Constructor.

        Arguments:
            call: System call to make
            kwargs: Keyword arguments passed to parent
        """
        self.call = call
        # caching on instance level as a compromise
        self._cache: dict[str, str] = {}  # {call: output}
        super().__init__(**kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(self, call=self.call)

    @property
    def output(self) -> str:
        key = " ".join(self.call)
        if key in self._cache:
            return self._cache[key]
        if output := helpers.get_output_from_call(self.call):
            self._cache[key] = output  # .replace("\n", "<br>").replace(" ", "&nbsp;")
            return self._cache[key]
        return "**Command failed**"

    def _to_markdown(self):
        call = " ".join(self.call)
        return TEXT.format(input=call, output=self.output)

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        if os.environ.get("CI"):
            node = MkCommandOutput(["make", "help"])
            node = mk.MkReprRawRendered(node)
        else:
            node = mk.MkHeader("Sadly breaks log output ATM, so only triggered in CI")
        page += node


if __name__ == "__main__":
    text = MkCommandOutput(["make", "help"])
    print(text)



from __future__ import annotations

import dataclasses

from mknodes.nodemods import modhelpers
from mknodes.utils import helpers, resources


SCROLLREVEAL_LINK = (
    "https://cdn.jsdelivr.net/npm/scrollreveal@3.4.0/dist/scrollreveal.min.js"
)
JQUERY_LINK = "https://code.jquery.com/jquery-2.2.4.min.js"

SCRIPT = """\
window.sr = ScrollReveal();

// Add class to <html> if ScrollReveal is supported
// Note: only works in version 3
if (sr.isSupported()) {
document.documentElement.classList.add('sr');
}
"""


SCRIPT_2 = """\
$(window).load(function(){

  window.sr = ScrollReveal();
  sr.reveal('.%s', %s);
});

"""

#   window.sr = ScrollReveal();
#   sr.reveal('.%s', {
#       origin: 'left',
#       distance: '300px',
#       easing: 'ease-in-out',
#       reset: true,
#       duration: 800,
#     });
# });


@dataclasses.dataclass(frozen=True)
class ScrollReveal:
    origin: str = "left"
    distance: str = "300px"
    easing: str = "ease-in-out"
    reset: bool = True
    duration: int = 800

    def get_resources(self):
        jquery_res = resources.JSFile(JQUERY_LINK, is_library=True)
        scrollreveal_res = resources.JSFile(SCROLLREVEAL_LINK, is_library=True)
        res1 = resources.JSText(SCRIPT, "scrollreveal.js")
        dct = dataclasses.asdict(self)
        js_map = modhelpers.format_js_map(dct)
        script_text = SCRIPT_2 % (self.css_class_names[-1], js_map)
        res2 = resources.JSText(script_text, "script.js")
        return resources.Resources(
            js=[
                jquery_res,
                scrollreveal_res,
                res1,
                res2,
            ],
        )

    @property
    def css_class_names(self):
        return ["sr", f"scrollreveal_{helpers.get_hash(self)}"]


if __name__ == "__main__":
    effect = ScrollReveal(origin="right")
    print(effect.get_resources().js[-1])



from __future__ import annotations

import pathlib

from git import TYPE_CHECKING

from mknodes.basenodes import mklink
from mknodes.navs import mknav, navbuilder
from mknodes.pages import mkpage


if TYPE_CHECKING:
    NavSubType = mknav.MkNav | mkpage.MkPage | mklink.MkLink


class Navigation(dict):
    """An object representing A Website structure.

    The dict data consists of a mapping of a path-tuple -> NavSubType.
    It can contain navs, which in turn have their own navigation object.

    A special item is the index page. It can be accessed via the corresponding attributes.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index_page: mkpage.MkPage | None = None
        self.index_title: str | None = None

    def __setitem__(self, index: tuple | str, node: NavSubType):
        """Put a Navigation-type instance into the registry.

        Index must be a unique path / title for this navigation object.
        If given a tuple, it is considered a nested path.
        """
        if isinstance(index, str):
            index = (index,)
        super().__setitem__(index, node)

    def __getitem__(self, index: tuple | str) -> NavSubType:
        if isinstance(index, str):
            index = (index,)
        return super().__getitem__(index)

    def __delitem__(self, index: tuple | str):
        if isinstance(index, str):
            index = (index,)
        super().__delitem__(index)

    def register(self, node: NavSubType):
        match node:
            case mknav.MkNav():
                self[(node.section,)] = node
            case mkpage.MkPage() | mklink.MkLink():
                self[node.title] = node
            case _:
                raise TypeError(node)

    @property
    def all_items(self):
        nodes = [self.index_page] if self.index_page else []
        nodes += list(self.values())
        return nodes

    @property
    def navs(self) -> list[mknav.MkNav]:
        """Return all registered navs."""
        return [node for node in self.values() if isinstance(node, mknav.MkNav)]

    @property
    def pages(self) -> list[mkpage.MkPage]:
        """Return all registered pages."""
        return [node for node in self.values() if isinstance(node, mkpage.MkPage)]

    @property
    def links(self) -> list[mklink.MkLink]:
        """Return all registered links."""
        return [node for node in self.values() if isinstance(node, mklink.MkLink)]

    def to_literate_nav(self) -> str:
        nav = navbuilder.NavBuilder()
        if self.index_page and self.index_title:
            nav[self.index_title] = pathlib.Path(self.index_page.path).as_posix()
        for path, item in self.items():
            if path is None:  # this check is just to make mypy happy
                continue
            match item:
                case mkpage.MkPage():
                    nav[path] = pathlib.Path(item.path).as_posix()
                case mknav.MkNav():
                    nav[path] = f"{item.section}/"
                case mklink.MkLink():
                    nav[path] = str(item.target)
                case _:
                    raise TypeError(item)
        return "".join(nav.build_literate_nav())



"""Classes and helpers for trees."""

from __future__ import annotations

from .node import Node
from .filetreenode import FileTreeNode
from .modulenode import ModuleNode
from .noderesolver import NodeResolver

__all__ = ["Node", "FileTreeNode", "ModuleNode", "NodeResolver"]



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from mknodes.basenodes import mklink
from mknodes.navs import mknav
from mknodes.pages import mkpage
from mknodes.utils import log


if TYPE_CHECKING:
    NavSubType = mknav.MkNav | mkpage.MkPage | mklink.MkLink

logger = log.get_logger(__name__)


class NavRouter:
    """Class used for MkNav decorator routing."""

    def __init__(self, nav: mknav.MkNav):
        """Constructor.

        Arguments:
            nav: MkNav to use for routing
        """
        self._nav = nav

    def __call__(
        self,
        *path: str,
    ) -> Callable[[Callable], Callable]:
        """Decorator method to use for routing.

        The decorated functions need to return either a MkPage or an MkNav.
        They will get registered to the router-MkNav then.

        In comparison to `page()` and `nav()`, it is not required to know in advance what
        will be returned.

        Arguments:
            path: The section path for the returned `MkNav` / `MkPage`
        """

        def decorator(fn: Callable[..., NavSubType], path=path) -> Callable:
            node = fn()
            node.parent = self._nav
            if isinstance(node, mkpage.MkPage):
                node.created_by = fn
            elif isinstance(node, mknav.MkNav) and node.index_page:
                node.index_page.created_by = fn
            self._nav.nav[path] = node
            return fn

        return decorator

    def nav(
        self,
        *path: str,
        **kwargs: Any,
    ) -> Callable[[Callable], Callable]:
        """Decorator method to use for routing Navs.

        The decorated functions will get passed an MkNav as an argument which can be
        modified then.
        They will get registered to the router-MkNav afterwards.

        There is also the possibility to create a new MkNav and return it in the
        decorated method, that MkNav takes preference then over the modified one.

        Examples:
            ``` py
            @nav.route.nav("Routed nav")
            def _(nav: mk.MkNav):
                nav += ...
            ```

        Arguments:
            path: The section path for the returned MkNav
            kwargs: Keyword arguments passed to the MkNav constructor.
        """

        def decorator(
            fn: Callable[..., mknav.MkNav],
            path=path,
            kwargs=kwargs,
        ) -> Callable:
            node = mknav.MkNav(path[-1], parent=self._nav, **kwargs)
            node = fn(node) or node
            node.parent = self._nav  # in case a new MkPage was generated
            if node.index_page:
                node.index_page.created_by = fn
            self._nav.nav[path] = node
            return fn

        return decorator

    def page(
        self,
        *path: str,
        **kwargs: Any,
    ) -> Callable[[Callable], Callable]:
        """Decorator method to use for routing Pages.

        The decorated functions will get passed an MkPage as an argument which can be
        modified then.
        The resulting page will get registered to the router-MkNav afterwards.

        There is also the possibility to create a new MkPage and return it in the
        decorated method, that MkPage takes preference then over the modified one.

        The keyword arguments can be used to pass metadata-related keyword arguments
        to the MkPage constructor.

        Examples:
            ``` py
            @nav.route.page("Routed page", icon=...)
            def _(page: mk.MkPage):
                page += ...
            ```
        Arguments:
            path: The section path for the returned MkPage
            kwargs: Keyword arguments passed to the MkPage constructor.
        """

        def decorator(
            fn: Callable[..., mkpage.MkPage],
            path=path,
            kwargs=kwargs,
        ) -> Callable:
            p = path[-1] if path else (self._nav.section or "Home")
            node = mkpage.MkPage(title=p, parent=self._nav, **kwargs)
            node = fn(node) or node
            node.parent = self._nav  # in case a new MkPage was generated
            node.created_by = fn
            self._nav.nav[path or self._nav.section or "Home"] = node
            return fn

        return decorator

    def link(self, *path: str, url: str, title: str):
        link = mklink.MkLink(url, title)
        self._nav[path] = link


if __name__ == "__main__":
    nav = mknav.MkNav()
    router = NavRouter(nav)



from __future__ import annotations

import types

from typing import TYPE_CHECKING, Any

from mknodes.basenodes import mknode
from mknodes.utils import icons, log, reprhelpers


if TYPE_CHECKING:
    import mknodes as mk

logger = log.get_logger(__name__)


class MkLink(mknode.MkNode):
    """A simple Link (with optional icon and option to show up as a button).

    If no title is given, the URL is used as a title.
    """

    ICON = "octicons/link-24"
    ATTR_LIST_SEPARATOR = ""

    def __init__(
        self,
        target: str | mk.MkPage | mk.MkNav | type | types.ModuleType,
        title: str | None = None,
        *,
        tooltip: str | None = None,
        icon: str | None = None,
        as_button: bool = False,
        primary_color: bool = False,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            target: Link target
            title: Title used for link
            tooltip: Tooltip for the link
            icon: Optional icon to be displayed in front of title
            as_button: Whether link should be rendered as button
            primary_color: If rendered as button, use primary color as background.
            kwargs: keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.target = target
        self._title = title
        self.tooltip = tooltip
        self.as_button = as_button
        self.primary_color = primary_color
        self._icon = icon or ""
        if as_button:
            self.add_css_class("md-button")
        if primary_color:
            self.add_css_class("md-button--primary")

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            target=self.target,
            title=self._title,
            tooltip=self.tooltip,
            icon=self._icon,
            as_button=self.as_button,
            primary_color=self.primary_color,
            _filter_empty=True,
            _filter_false=True,
        )

    @property
    def icon(self) -> str:
        if not self._icon:
            return ""
        return icons.get_emoji_slug(self._icon)

    @property
    def url(self) -> str:
        return self.ctx.links.get_url(self.target)

    @property
    def title(self) -> str:
        return self.url if self._title is None else self._title

    def _to_markdown(self) -> str:
        prefix = f"{self.icon} " if self.icon else ""
        tooltip = f" {self.tooltip!r}" if self.tooltip else ""
        return f"[{prefix}{self.title}]({self.url}{tooltip})"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        url = "http://www.google.de"
        node = mk.MkLink(url, "This is a link.")
        page += mk.MkReprRawRendered(node, header="### Regular")
        node = mk.MkLink(url, "Disguised as button.", as_button=True)
        page += mk.MkReprRawRendered(node, header="### Button")
        node = mk.MkLink(url, "Colored.", as_button=True, primary_color=True)
        page += mk.MkReprRawRendered(node, header="### Colored")
        node = mk.MkLink(url, "With icon.", icon="octicons/link-24")
        page += mk.MkReprRawRendered(node, header="### With icon")
        node = mk.MkLink(url, "With tooltip.", tooltip="Test")
        page += mk.MkReprRawRendered(node, header="### With tooltip")
        node = mk.MkLink(page.parent.index_page, "To page.")
        page += mk.MkReprRawRendered(node, header="###To page")


if __name__ == "__main__":
    link = MkLink("www.test.de")



from __future__ import annotations

from typing import Any
from xml.etree import ElementTree as Et

from mknodes.utils import icons, log


logger = log.get_logger(__name__)


class HTMLElement(Et.Element):
    tag_name: str

    def __init__(
        self,
        klass: str | None = None,
        parent: Et.Element | None = None,
        *,
        attrs: dict | None = None,
        markdown: bool = False,
        **kwargs: str,
    ):
        """Constructor.

        Arguments:
            klass: CSS class
            parent: Optional parent element
            attrs: A dict containing XML attributes
            markdown: Set markdown attribute. (Also sets markdown attr for parents)
            kwargs: Additional XML attributes
        """
        kls = {"class": klass} if klass else {}
        self.parent = parent
        attrs = attrs or {}
        super().__init__(self.tag_name, attrs | kwargs | kls)
        if parent is not None:
            parent.append(self)
        if markdown:
            node = self
            while node is not None:
                node.set("markdown", "1")
                if isinstance(node, HTMLElement) and node.parent:
                    node = node.parent
                else:
                    break

    def to_string(self, space: str = "  ", level: int = 0) -> str:
        Et.indent(self, space=space, level=level)
        return Et.tostring(self, encoding="unicode", method="html")


class Div(HTMLElement):
    tag_name = "div"

    def __init__(
        self,
        klass: str | None = None,
        text: str | None = None,
        parent: Et.Element | None = None,
        **kwargs: Any,
    ):
        super().__init__(klass, parent, **kwargs)
        self.text = text


class Header(HTMLElement):
    def __init__(
        self,
        level: int,
        text: str,
        parent: Et.Element | None = None,
        **kwargs: Any,
    ):
        self.tag_name = f"h{level}"
        super().__init__(**kwargs)
        self.text = text
        if parent is not None:
            parent.append(self)


class P(HTMLElement):
    tag_name = "p"


class Section(HTMLElement):
    tag_name = "section"


class Span(HTMLElement):
    tag_name = "span"

    def __init__(
        self,
        klass: str | None = None,
        text: str | None = None,
        parent: Et.Element | None = None,
        **kwargs: Any,
    ):
        super().__init__(klass, parent, **kwargs)
        self.text = text


class Img(HTMLElement):
    tag_name = "img"


class Ul(HTMLElement):
    tag_name = "ul"


class Li(HTMLElement):
    tag_name = "li"


class Table(HTMLElement):
    tag_name = "table"


class Tr(HTMLElement):
    tag_name = "tr"


class Aside(HTMLElement):
    tag_name = "aside"


class Td(HTMLElement):
    tag_name = "td"


class Button(HTMLElement):
    tag_name = "button"


class Form(HTMLElement):
    tag_name = "form"


class Input(HTMLElement):
    tag_name = "input"


class Label(HTMLElement):
    tag_name = "label"


class A(HTMLElement):
    tag_name = "a"

    def __init__(
        self,
        klass: str | None = None,
        text: str | None = None,
        parent: Et.Element | None = None,
        **kwargs: Any,
    ):
        super().__init__(klass, parent, **kwargs)
        self.text = text


def get_source_button(icon: str = "material/code-json") -> A:
    href = (
        "{{ config.site_url | rstrip('/') + '/src/' + page.file.src_uri | replace('.md',"
        " '.original') }}"
    )
    elem = A(
        "md-content__button md-icon",
        title="Source",
        href=href,
    )
    svg = icons.get_icon_xml(icon)
    elem.append(svg)
    return elem


def pformat(str_or_elem: str | Et.Element, space: str = "  ", level: int = 0) -> str:
    """Prettyprint given XML.

    Arguments:
        str_or_elem: XML to prettyprint
        space: Amount of spaces to use for indentation
        level: Initial indentation level
    """
    if isinstance(str_or_elem, str):
        str_or_elem = Et.fromstring(str_or_elem)
    Et.indent(str_or_elem, space=space, level=level)
    return Et.tostring(str_or_elem, encoding="unicode", method="html")


if __name__ == "__main__":
    a = Div("test")



from __future__ import annotations

from typing import Any

from mknodes.basenodes import mknode
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


class MkKeys(mknode.MkNode):
    """Node to display Keyboard keys.

    Mostly used to explain shortcuts / commands for the keyboard.
    """

    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.keys")]
    ICON = "fontawesome/regular/keyboard"
    ATTR_LIST_SEPARATOR = ""

    def __init__(
        self,
        keys: str | list[str],
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            keys: keys to display
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        match keys:
            case str():
                self.keys = keys.lower().split("+")
            case list():
                self.keys = [i.lower() for i in keys]

    def __repr__(self):
        return reprhelpers.get_repr(self, keys=self.keys)

    def _to_markdown(self) -> str:
        key_str = "+".join(self.keys)
        return f"++{key_str}++"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        page += mk.MkAdmonition("MkKeys can be used to show Key combinations.")
        page += mk.MkReprRawRendered(MkKeys("M+k+K+e+y+s"))
        page += mk.MkReprRawRendered(MkKeys("Ctrl+A"))


if __name__ == "__main__":
    keys = MkKeys(keys="Ctrl+A")
    keys.add_css_class("test")
    print(keys.to_html())



from __future__ import annotations

import os
import pathlib
import textwrap

from typing import Any, Self

from mknodes.basenodes import mknode
from mknodes.data import datatypes
from mknodes.utils import classhelpers, inspecthelpers, log, reprhelpers, richhelpers


logger = log.get_logger(__name__)


class MkCodeImage(mknode.MkNode):
    """Node to display a code block as an SVG image."""

    ICON = "material/code-json"

    def __init__(
        self,
        code: str | datatypes.HasCodeType = "",
        language: str = "py",
        *,
        title: str = "",
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            code: Code to show
            language: language for syntax highlighting
            title: Code block title
            kwargs: Keyword arguments passed to parent
        """
        self.language = language
        self.title = title
        self._code = code
        super().__init__(**kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            code=self._code,
            language=self.language,
            title=self.title,
            _filter_empty=True,
        )

    @property
    def code(self) -> str:
        match self._code:
            case str():
                return self._code
            case _:
                return inspecthelpers.get_source(self._code)

    def _to_markdown(self) -> str:
        content = richhelpers.get_svg_for_code(
            self.code,
            language=self.language,
            title=self.title,
        )
        return f"<body>\n\n{content}\n\n</body>\n"

    @classmethod
    def create_example_page(cls, page):
        node = MkCodeImage(language="python", code="a = 1 + 2", title="Window title")
        page += node

    @classmethod
    def for_file(
        cls,
        path: str | os.PathLike,
        *,
        title: str | None = None,
        **kwargs: Any,
    ):
        """Create a MkCode node based on a code file.

        Arguments:
            path: Path to the code file
            title: title to use for the code box. If None is set, filename will be used.
            kwargs: Keyword arguments passed to MkCode ctor
        """
        path = pathlib.Path(path)
        with path.open() as file:
            content = file.read()
        title = path.name if title is None else title
        return cls(content, title=title, **kwargs)

    @classmethod
    def for_object(
        cls,
        obj: datatypes.HasCodeType,
        *,
        dedent: bool = True,
        extract_body: bool = False,
        title: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a MkCode node based on a python object.

        Fetches code by using the inspect module.

        Arguments:
            obj: Python object to show code from
            dedent: Whether to dedent the code
            extract_body: if True, Function / Class signatures are stripped from the code
            title: Title to use for code block. If None, it will use the object path.
            kwargs: Keyword arguments passed to MkCode ctor
        """
        if extract_body and callable(obj):
            code = inspecthelpers.get_function_body(obj)
        elif extract_body:
            msg = "Can only extract body from Functions, Methods and classes"
            raise TypeError(msg)
        else:
            code = inspecthelpers.get_source(obj)
        code = textwrap.dedent(code) if dedent else code
        code_title = title if title is not None else classhelpers.get_code_name(obj)
        return cls(code=code, title=code_title, **kwargs)


if __name__ == "__main__":
    node = MkCodeImage(language="python", code="a = 1 + 2")
    print(node)



from __future__ import annotations

from mknodes.nodemods import parallax, scrollreveal
from mknodes.utils import log, resources


logger = log.get_logger(__name__)


class ModManager:
    def __init__(self):
        self.mods = []
        self._css_classes = []

    def __hash__(self):
        return sum([hash(i) for i in self.css_classes])

    def __eq__(self, other):
        """Needed for Mknode comparison."""
        return hash(self) == hash(other)

    def append(self, other):
        if isinstance(other, str):
            self._css_classes.append(other)
        else:
            self.mods.append(other)

    def get_resources(self) -> resources.Resources:
        req = resources.Resources()
        for mod in self.mods:
            req.merge(mod.get_resources())
        return req

    @property
    def css_classes(self) -> list[str]:
        cls_names = [cls_name for mod in self.mods for cls_name in mod.css_class_names]
        return self._css_classes + cls_names

    @property
    def attr_list_str(self):
        classes = " ".join(self.css_classes)
        return f"{{: .{classes}}}"

    def add_parallax_effect(self, orientation: str = "up"):
        effect = parallax.ParallaxEffect(orientation=orientation)
        self.mods.append(effect)

    def add_scroll_effect(self, origin: str = "left"):
        effect = scrollreveal.ScrollReveal(origin=origin)
        self.mods.append(effect)


if __name__ == "__main__":
    manager = ModManager()
    print(manager.get_resources())



from __future__ import annotations

import itertools

from typing import Any, Literal

from mknodes.basenodes import mkdiagram
from mknodes.utils import classhelpers, connector, helpers, reprhelpers


DiagramModeStr = Literal["baseclasses", "subclasses", "mro"]


class BaseClassConnector(connector.Connector):
    def __init__(
        self,
        objects,
        *,
        title_style: Literal["package.classname", "qualname"] = "package.classname",
        max_depth: int | None = None,
    ):
        self.title_style = title_style
        # self.object = objects[0]
        super().__init__(objects, max_depth=max_depth)

    def get_id(self, item: type) -> int:
        return id(item)

    def get_title(self, item: type) -> str:
        return (
            helpers.label_for_class(item)
            if self.title_style == "package.classname"
            else item.__qualname__
        )
        # if item.__module__.split(".")[0] == self.object.__module__.split(".")[0]:
        #     return f"**{text}**"
        # else:
        #     return text

    def get_attributes(self, item) -> list[str]:
        return [i for i in dir(item) if not i.startswith("__")]


class SubclassConnector(BaseClassConnector):
    def _connect(self, objects):
        super()._connect(objects)
        self.connections = [(i[1], i[0]) for i in self.connections]

    def get_children(self, item: type) -> list[type]:
        return item.__subclasses__()


class ParentClassConnector(BaseClassConnector):
    def get_children(self, item: type) -> tuple[type, ...]:
        return item.__bases__


class MroConnector(BaseClassConnector):
    def _connect(self, objects):
        mro = list(objects[0].mro())[: self.max_depth]
        self.item_dict = {self.get_id(kls): self.get_title(kls) for kls in mro}
        self.connections = [
            (self.get_id(i), self.get_id(j)) for i, j in itertools.pairwise(mro)
        ]


class MkClassDiagram(mkdiagram.MkDiagram):
    """Node to display the class hierarchy of a class. Supports multiple modes."""

    def __init__(
        self,
        klass: type | str,
        mode: DiagramModeStr = "baseclasses",
        direction: Literal["TD", "DT", "LR", "RL"] = "TD",
        max_depth: int | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            klass: Klass to show a diagram for
            mode: What kind of diagram to display
            direction: diagram direction
            max_depth: Maxium depth to display
            kwargs: Keyword arguments passed to parent
        """
        self.klass = klass
        self.mode = mode
        self._max_depth = max_depth
        super().__init__(graph_type="flow", direction=direction, **kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            klass=self.klass,
            mode=self.mode,
            direction=self.direction,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        parent_diagram = MkClassDiagram(klass=MkClassDiagram, mode="baseclasses")
        sub = MkClassDiagram(klass=mk.MkContainer, mode="subclasses", direction="LR")
        mro_diagram = MkClassDiagram(klass=mk.MkTable, mode="mro")
        page += "### Parent class hierarchy: MkClassDiagram"
        page += mk.MkReprRawRendered(parent_diagram)
        page += "### Subclass hierarchy: MkContainer"
        page += mk.MkReprRawRendered(sub)
        page += "### Method resolution order: MkTable"
        page += mk.MkReprRawRendered(mro_diagram)

    @property
    def mermaid_code(self) -> str:
        klass = classhelpers.to_class(self.klass)
        match self.mode:
            case "subclasses":
                builder = SubclassConnector(klass, max_depth=self._max_depth)
                return builder.get_graph_connection_text()
            case "baseclasses":
                builder = ParentClassConnector(klass, max_depth=self._max_depth)
                return builder.get_graph_connection_text()
            case "mro":
                builder = MroConnector(klass, max_depth=self._max_depth)
                return builder.get_graph_connection_text()
            case _:
                raise ValueError(self.mode)


if __name__ == "__main__":
    import mknodes as mk

    diagram = MkClassDiagram(mk.MkNode, mode="subclasses", max_depth=3)
    print(diagram)



from __future__ import annotations

from collections.abc import Callable
import inspect
import types

from typing import Any

from mknodes.treelib import node
from mknodes.utils import inspecthelpers, log


logger = log.get_logger(__name__)


class ModuleNode(node.Node):
    def __init__(self, module: types.ModuleType, **kwargs: Any):
        self.module = module
        self.summary = inspecthelpers.get_doc(self.module, only_summary=True)
        super().__init__(**kwargs)

    def __repr__(self):
        module_name = self.module.__name__.split(".")[-1]
        if self.summary:
            module_name += f": {self.summary}"
        return module_name

    @classmethod
    def from_module(
        cls,
        module: types.ModuleType,
        *,
        predicate: Callable | None = None,
        exclude: list[str] | None = None,
        sort: bool = True,
        max_items: int | None = None,
        maximum_depth: int | None = None,
        parent: ModuleNode | None = None,
        _seen=None,
    ):
        node = cls(module, parent=parent)
        seen = _seen or set()
        children = [
            mod
            for _name, mod in inspect.getmembers(module, inspect.ismodule)
            if mod.__name__.startswith(module.__name__)
        ]
        if sort:
            children = sorted(children, key=lambda s: s.__name__.lower())
        for submod in children:
            if predicate and not predicate(submod):
                continue
            if exclude and submod.__name__ in exclude:
                continue
            if not children:
                child = cls(submod, parent=node)
            else:
                if maximum_depth is not None and maximum_depth < node.depth + 1:
                    continue
                child = None
                if submod not in seen:
                    seen.add(submod)
                    child = ModuleNode.from_module(
                        submod,
                        parent=node,
                        predicate=predicate,
                        max_items=max_items,
                        maximum_depth=maximum_depth,
                        exclude=exclude,
                        _seen=seen,
                    )
            if max_items is not None:
                if max_items > 0:
                    max_items -= 1
                else:
                    break
            if child:
                node.append_child(child)
        return node


if __name__ == "__main__":
    import mknodes as mk

    folder = ModuleNode.from_module(mk)
    logger.warning(folder.get_tree_repr())
    # node = ModuleNode(mk)
    # print(node.children)
    # logger.warning(node.get_tree_repr())



from __future__ import annotations

import pathlib

from typing import TYPE_CHECKING, Any

import jinja2

from mknodes.jinja import environment
from mknodes.utils import inspecthelpers, log


if TYPE_CHECKING:
    import mknodes as mk


logger = log.get_logger(__name__)


class NodeEnvironment(environment.Environment):
    """Jinja Node environment.

    A Jinja Environment specifically for MkNode instances.

    - Sets the parent for the filters
    - Puts node context in jinja namespace
    - collects rendered nodes
    """

    def __init__(self, node: mk.MkNode, **kwargs: Any):
        """Constructor.

        Arguments:
            node: Node this environment belongs to.
            kwargs: Optional keyword arguments passed to parent
        """
        super().__init__(load_templates=True, **kwargs)
        self.inherit_from(node.ctx.env)
        self.node = node
        self.rendered_nodes: list[mk.MkNode] = list()
        self.setup_environment()
        path = inspecthelpers.get_file(self.node.__class__)  # type: ignore[arg-type]
        self.class_path = pathlib.Path(path or "").parent.as_posix()
        paths = self.get_extra_paths()
        self.add_template_path(*paths)

    def setup_environment(self):
        import mknodes as mk

        filters = {}
        for kls_name in mk.__all__:

            def wrapped(ctx, *args, kls_name=kls_name, **kwargs):
                kls = getattr(mk, kls_name)
                node = kls(*args, parent=self.node, **kwargs)
                self.rendered_nodes.append(node)
                return node

            filters[kls_name] = jinja2.pass_context(wrapped)
        self.filters.update(filters)
        self.globals["parent_page"] = self.node.parent_page
        self.globals["parent_nav"] = (
            self.node.parent_navs[-1] if self.node.parent_navs else None
        )
        self.globals["mknode"] = self.node
        self.globals["mk"] = filters

    def update_env_from_context(self):
        self.filters["get_link"] = self.node.ctx.links.get_link
        self.filters["get_url"] = self.node.ctx.links.get_url
        self.globals |= self.node.ctx.as_dict()

    def get_extra_paths(self) -> list[str]:
        paths = [self.class_path]
        if self.node.parent_navs:
            nav = self.node.parent_navs[-1]
            if "created" in nav.metadata:
                file = nav.metadata["created"]["source_filename"]
                path = pathlib.Path(file).parent
                paths.append(path.as_posix())
        return paths

    def render_template(
        self,
        template_name: str,
        variables: dict[str, Any] | None = None,
        block_name: str | None = None,
        parent_template: str | None = None,
    ) -> str:
        """Render a loaded template.

        Rendered nodes can be collected from `rendered_nodes` attribute after this call.

        Arguments:
            template_name: Template name
            variables: Extra variables for this render call
            block_name: Render specific block from the template
            parent_template: Optional parent template (to be used with super())
        """
        # if pathlib.Path(template_name).as_posix() not in self.list_templates():
        #     self.add_template(template_name)
        self.rendered_nodes = []
        self.update_env_from_context()
        return super().render_template(
            template_name,
            variables=variables,
            block_name=block_name,
            parent_template=parent_template,
        )

    def render_string(self, markdown: str, variables: dict | None = None):
        """Render a template string.

        Rendered nodes can be collected from `rendered_nodes` attribute after this call.

        Arguments:
            markdown: String to render
            variables: Extra variables for the environment
        """
        self.rendered_nodes = []
        self.update_env_from_context()
        return super().render_string(markdown, variables)

    def set_mknodes_filters(self):
        """Set our MkNode filters."""
        import mknodes as mk

        filters = {}
        for kls_name in mk.__all__:

            def wrapped(ctx, *args, kls_name=kls_name, **kwargs):
                kls = getattr(mk, kls_name)
                node = kls(*args, parent=self.node, **kwargs)
                self.rendered_nodes.append(node)
                return node

            filters[kls_name] = jinja2.pass_context(wrapped)
        self.filters.update(filters)
        self.globals["mk"] = filters


if __name__ == "__main__":
    import mknodes as mk

    node = mk.MkText.with_default_context()
    env = NodeEnvironment(node)
    txt = "{{ metadata.required_python_version | MkAdmonition }}"
    print(env.render_string(txt))
    print(env.rendered_nodes)
    # text = env.render_string(r"{{ 'test' | MkHeader }}")
    # text = env.render_string(r"{{ 50 | MkProgressBar }}")
    # print(env.rendered_nodes)
    # env.render_string(r"{{test('hallo')}}")



import mknodes as mk


nav = mk.MkNav("MkPages")


@nav.route.page(is_index=True)
def _(page: mk.MkPage):
    page += mk.MkJinjaTemplate("pages/mkpage_index.jinja")
    page.footnotes[1] = r"Footnotes are numbered, can be set via \__setitem__."
    page.footnotes[2] = r"They can also get nested[^3]."
    page.footnotes[3] = mk.MkAdmonition("They can also contain other Markdown (nodes).")


@nav.route.page("MkClassPage")
def _(page: mk.MkPage):
    variables = dict(example_class=mk.MkCode)
    page += mk.MkJinjaTemplate("pages/mkclasspage.jinja", variables=variables)


@nav.route.page("MkModulePage")
def _(page: mk.MkPage):
    import mkdocs.config

    variables = dict(example_module=mkdocs.config)
    page += mk.MkJinjaTemplate("pages/mkmodulepage.jinja", variables=variables)


@nav.route.page("Setting the homepage")
def _(page: mk.MkPage):
    page += mk.MkJinjaTemplate("pages/homepage.jinja")


@nav.route.page("Adding to MkPages", hide="toc", status="new")
def _(page: mk.MkPage):
    page += mk.MkAdmonition("You can add other MkNodes to a page sequentially.")
    page += "Adding strings also works, they get converted to MkText nodes."
    page += "### ...and text starting with # will become a MkHeader."
    page += "Every MkPage has a MkFootNotes node built-in[^1]."
    page.footnotes[1] = "Super cool, right?"


@nav.route.page(
    "Metadata",
    status="deprecated",
    search_boost=2.0,
    subtitle="Subtitle",
    description="Description",
)
def _(page: mk.MkPage):
    page += mk.MkJinjaTemplate("pages/page_metadata.jinja")


@nav.route.page("Templates", hide="toc", status="new")
def _(page: mk.MkPage):
    page += mk.MkJinjaTemplate("pages/page_templates.jinja")
    page.template.announce.content = mk.MkMetadataBadges(typ="classifiers")
    page.template.footer.content = mk.MkProgressBar(50)
    code = "information = 'You can even put MkNodes here!'"
    page.template.tabs.content = mk.MkCode(f"{code}")
    page.template.hero.content = mk.MkHeader("A header!")
    css = {
        ":root": {
            "--md-primary-fg-color": "#FF0000",
            "--md-primary-fg-color--light": "#FF0000",
            "--md-primary-fg-color--dark": "#FF0000",
        },
    }
    page.template.styles.add_css(css)


@nav.route.page("Resources")
def _(page: mk.MkPage):
    page += mk.MkJinjaTemplate("resources.jinja")
